# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import itertools
import operator
import re
from collections import abc

from ...utils import cliparser

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


BOOLEAN = 'boolean'
COMPARATIVE = 'comparative'


class BoolFilterSpec():
    """Boolean filter specification"""

    type = BOOLEAN

    def __init__(self, func, *, needed_keys=(), aliases=(), description='No description'):
        if not func:
            self.filter_function = None
            needed_keys = ()
        else:
            self.filter_function = func
        self.needed_keys = needed_keys
        self.aliases = aliases
        self.description = description


class CmpFilterSpec():
    """Comparative filter specification"""

    type = COMPARATIVE

    def __init__(self, *, value_type, value_getter=None, value_matcher=None,
                 value_convert=None, as_bool=None, needed_keys=(), aliases=(),
                 description='No description'):
        """
        value_type    : Subclass of `type` (i.e. something that returns an instance when
                        called and can be passed to `isinstance` as the second argument
        value_getter  : Callable that takes an item and returns one or more
                        values to match against the user-provided value;
                        Multiple values must be given as an iterator (list,
                        tuple, generator, etc), and the item matches if any
                        match
        value_convert : Callable that takes a value and converts it to something
                        comparable (e.g. "42" (str) -> 42 (int))
        value_matcher : Callable that takes (item, operator, value) and returns True/False
        as_bool       : Callable that takes an item and returns True/False
        needed_keys   : Needed keys for this filter
        aliases       : Alternative names of this filter
        """
        self.value_type = value_type
        self.needed_keys = needed_keys
        self.aliases = aliases
        self.description = description
        self.value_convert = value_convert if value_convert is not None else value_type

        if value_getter is not None:
            self.value_getter = value_getter
        elif len(self.needed_keys) == 1:
            self.value_getter = lambda dct, k=needed_keys[0]: dct[k]
        else:
            raise TypeError('Missing argument with needed_keys=%r: value_getter', self.needed_keys)

        if value_matcher is None:
            def value_matcher(item, op, user_value, vg=self.value_getter):
                item_value = vg(item)
                if isinstance(item_value, abc.Iterator):
                    return any(op(ival, user_value) for ival in item_value)
                else:
                    return op(item_value, user_value)
        self.value_matcher = value_matcher

        if as_bool is None:
            def as_bool(item, vg=self.value_getter):
                item_value = vg(item)
                if isinstance(item_value, abc.Iterator):
                    return any(item_value)
                else:
                    return bool(item_value)
        self.as_bool = as_bool

    def make_filter(self, operator, user_value, invert):
        if operator is None and user_value is None:
            # Abuse comparative filter as boolean filter
            # (e.g. 'peers-connected' matches torrents with peers-connected!=0)
            return (self.as_bool, self.needed_keys, invert)
        elif user_value is None:
            # Operator with no value matches everything
            return (None, (), False)
        else:
            def f(obj, vm=self.value_matcher, op=operator, val=user_value):
                return vm(obj, op, val)
            return (f, self.needed_keys, invert)


class FilterSpecDict(abc.Mapping):
    """TODO"""
    _NOT_FOUND = object()

    def __init__(self, dct):
        self._dct = dct

    def __getitem__(self, key):
        value = self._dct.get(key, self._NOT_FOUND)
        if value is not self._NOT_FOUND:
            return value
        for value in self._dct.values():
            if key in value.aliases:
                return value
        raise KeyError(key)

    def __iter__(self):
        return iter(self._dct)

    def __len__(self):
        return len(self._dct)

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, self._dct)


class Filter():
    """Match sequences of objects against a single filter"""

    OPERATORS = {
        '='  : operator.__eq__, '~'  : operator.__contains__,
        '>'  : operator.__gt__, '<'  : operator.__lt__,
        '>=' : operator.__ge__, '<=' : operator.__le__,
        '=~' : lambda a, b: re.search(b, a),
    }
    INVERT_CHAR = '!'
    POSSIBLE_OPERATORS = tuple(itertools.chain.from_iterable((op, '!' + op)
                                                             for op in OPERATORS))
    DEFAULT_FILTER = None
    DEFAULT_OPERATOR = '~'
    BOOLEAN_FILTERS = {}
    COMPARATIVE_FILTERS = {}

    @classmethod
    def _resolve_alias(cls, name):
        """
        Return real filter name or `name` if it does not resolve
        """
        if not hasattr(cls, '_aliases'):
            aliases = {}
            for fspecs in (cls.BOOLEAN_FILTERS, cls.COMPARATIVE_FILTERS):
                for fname,f in fspecs.items():
                    for a in f.aliases:
                        if a in aliases:
                            raise RuntimeError('Multiple aliases: %r' % (a,))
                        else:
                            aliases[a] = fname
            cls._aliases = aliases
        if name is None:
            name = ''
        return cls._aliases.get(name.strip(), name)

    @classmethod
    def _get_filter_spec(cls, name):
        """
        Get filter spec by `name`

        Raise ValueError on error
        """
        fspec = cls.BOOLEAN_FILTERS.get(name)
        if fspec is not None:
            return fspec
        fspec = cls.COMPARATIVE_FILTERS.get(name)
        if fspec is not None:
            return fspec
        if name:
            raise ValueError('Invalid filter name: %r' % (name,))
        else:
            raise ValueError('No filter expression given')

    @classmethod
    def _make_filter(cls, name, op, user_value, invert):
        """
        Return filter function, needed keys and invert

        Filter function takes a value and returns whether it matches
        `user_value`.

        Filter function and needed keys are both `None` if everything is
        matched.

        Raise ValueError on error
        """
        # Ensure value is wanted by filter, compatible to operator and of proper type
        user_value = cls._validate_user_value(name, op, user_value)
        log.debug('  Validated user_value: %r', user_value)

        fspec = cls._get_filter_spec(name)
        if fspec.type is BOOLEAN:
            return (fspec.filter_function, fspec.needed_keys, invert)
        elif fspec.type is COMPARATIVE:
            return fspec.make_filter(cls.OPERATORS.get(op), user_value, invert)

    @classmethod
    def _validate_user_value(cls, name, op, user_value):
        """
        Ensure that the `name`, `op`, and `user_value` make sense in conjunction

        Return user value as correct type (e.g. `int`) for filter `name`

        Raise ValueError if anything smells funky
        """
        log.debug('  Validating user value: name=%r, op=%r, user_value=%r',
                  name, op, user_value)

        if name in cls.BOOLEAN_FILTERS:
            # log.debug('%r is a valid boolean filter: %r', name, cls.BOOLEAN_FILTERS[name])
            if user_value:
                raise ValueError('Boolean filter does not take a value: %s' % (name,))
            elif op:
                raise ValueError('Boolean filter does not take an operator: %s' % (name,))

        if op is None or user_value is None:
            # Filter `name` could still be (ab)used as boolean filter
            return None

        fspec = cls.COMPARATIVE_FILTERS.get(name)
        if fspec is None:
            if name:
                raise ValueError('Invalid filter name: %r' % (name,))
            else:
                raise ValueError('No filter expression given')

        # Convert user_value to proper type
        if type(user_value) is not fspec.value_type:
            log.debug('  Converting %r to %r', user_value, fspec.value_type)
            try:
                user_value = fspec.value_convert(user_value)
            except ValueError:
                raise ValueError('Invalid value for filter %r: %r' % (name, user_value))

        # In case of regex operator, compile user_value
        if op == '=~':
            try:
                user_value = re.compile(user_value)
            except re.error as e:
                raise ValueError('Invalid regular expression: %s: %s' % (str(e).capitalize(), user_value))
        else:
            # Test if target_type supports operator
            try:
                log.debug('Trying %r(%r [%r], %r [%r])',
                          cls.OPERATORS[op], user_value, type(user_value), user_value, type(user_value))
                cls.OPERATORS[op](user_value, user_value)
            except TypeError:
                raise ValueError('Invalid operator for filter %r: %s' % (name, op))

        return user_value

    @classmethod
    def _parse_inverter(cls, string, invert):
        if not string:
            return string, invert

        # Find INVERT_CHAR at start or end of string
        parts = cliparser.tokenize(string.strip(), delims=(cls.INVERT_CHAR,), escapes=('\\',), quotes=())
        if cls.INVERT_CHAR in parts:
            if parts and parts[0] == cls.INVERT_CHAR:
                parts.pop(0)
                invert = not invert
            if parts and parts[-1] == cls.INVERT_CHAR:
                parts.pop(-1)
                invert = not invert
            return ''.join(parts), invert
        else:
            # Return string unchanged
            return string, invert

    def __init__(self, filter_str=''):
        # name: Name of filter (user-readable string)
        # invert: Whether to invert filter (bool)
        # op: Comparison operator as string (see OPERATORS)
        # user_value: User-given value that is matched against items
        # The *_raw variables contain original quotes and backslashes.
        name_raw, op_raw, user_value_raw, invert = (None, None, None, False)

        log.debug('Parsing %r', filter_str)
        parts = cliparser.tokenize(filter_str, maxdelims=1, delims=self.OPERATORS, escapes=('\\',))
        log.debug('Parts: %r', parts)
        if len(parts) == 3:
            name_raw, op_raw, user_value_raw = parts
        elif len(parts) == 2:
            if parts[0] in self.OPERATORS:
                op_raw, user_value_raw = parts
                name_raw = self.DEFAULT_FILTER
            elif parts[1] in self.OPERATORS:
                name_raw, op_raw = parts
            else:
                raise ValueError('Malformed filter expression: %r' % (filter_str,))
        elif len(parts) == 1:
            if parts[0] in self.OPERATORS:
                op_raw = parts[0]
            else:
                name_raw = parts[0]
        else:
            raise ValueError('Malformed filter expression: %r' % (filter_str,))
        name_raw, invert = self._parse_inverter(name_raw, invert)
        log.debug('Parsed %r into raw: name=%r, invert=%r, op=%r, user_value=%r',
                  filter_str, name_raw, invert, op_raw, user_value_raw)

        # Remove all special characters (backslashes, quotes)
        name, op, user_value = map(lambda x: None if x is None else cliparser.plaintext(x),
                                   (name_raw, op_raw, user_value_raw))
        log.debug('  Plaintext: name=%r, invert=%r, op=%r, user_value=%r',
                  name, invert, op, user_value)

        name = self._resolve_alias(name)
        log.debug('  Resolved alias: name=%r, op=%r, user_value=%r', name, op, user_value)

        if not name:
            name = self.DEFAULT_FILTER
            log.debug('  Falling back to default filter: %r', name)

        try:
            log.debug('  Getting filter spec: name=%r, op=%r, user_value=%r', name, op, user_value)
            # Get filter spec by `name`
            filter_func, needed_keys, invert = self._make_filter(name, op, user_value, invert)
        except ValueError:
            # Filter spec lookup failed
            if self.DEFAULT_FILTER and user_value is op is None:
                # No `user_value` or `op` given - use the first part of the
                # filter expression (normally the filter name) as `user_value`
                # for DEFAULT_FILTER.
                name, op, user_value = self.DEFAULT_FILTER, self.DEFAULT_OPERATOR, name
                log.debug('  Using name as value for default filter: name=%r, op=%r, user_value=%r',
                          name, op, user_value)
                filter_func, needed_keys, invert = self._make_filter(name, op, user_value, invert)
            else:
                # No DEFAULT_FILTER is set, so we can't default to it
                raise

        log.debug('  Final filter: name=%r, invert=%r, op=%r, user_value=%r',
                  name, invert, op, user_value)
        self._filter_func = filter_func
        self._needed_keys = needed_keys
        self._name, self._invert, self._op, self._user_value = name, invert, op, user_value
        self._hash = hash((name, invert, op, user_value))

    def apply(self, objs, invert=False, key=None):
        """Yield matching objects or `key` of each matching object"""
        invert = self._invert ^ bool(invert)  # xor
        is_wanted = self._filter_func
        if is_wanted is None:
            if invert:
                # This filter matches nothing
                yield from ()
            else:
                # This filter matches everything
                if key is None:
                    yield from objs
                else:
                    for obj in objs:
                        yield obj[key]
        else:
            if key is None:
                for obj in objs:
                    if bool(is_wanted(obj)) ^ invert:
                        yield obj
            else:
                for obj in objs:
                    if bool(is_wanted(obj)) ^ invert:
                        yield obj[key]

    def match(self, obj):
        """Return True if `obj` matches, False otherwise"""
        is_wanted = self._filter_func
        if is_wanted is None:
            # This filter matches everything/nothing
            return not self._invert
        else:
            return bool(is_wanted(obj)) ^ self._invert

    def __str__(self):
        if self._name is None:
            return self.DEFAULT_FILTER or ''
        elif self._op is None:
            return ('!' if self._invert else '') + self._name
        else:
            name = self._name if self._name != self.DEFAULT_FILTER else ''
            op = ('!' if self._invert else '') + self._op
            user_value = self._user_value
            if user_value is None:
                return name + op
            else:
                val = str(user_value)
                if val == '':
                    val = "''"
                elif len(val) == 1:
                    val = cliparser.escape(val, delims=(' ', '&', '|'), quotes=("'", '"'))
                else:
                    val = cliparser.quote(val, delims=(' ', '&', '|'), quotes=("'", '"'))
                return name + op + val

    @property
    def needed_keys(self):
        return self._needed_keys

    @property
    def match_everything(self):
        return not self._filter_func

    @property
    def inverted(self):
        return self._invert

    def __eq__(self, other):
        if isinstance(other, type(self)):
            for attr in ('_name', '_user_value', '_invert', '_op'):
                if getattr(self, attr) != getattr(other, attr):
                    return False
            return True
        else:
            return NotImplemented

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, str(self))

    def __hash__(self):
        return self._hash


# The filter specs are specified on the Filter subclasses in each module, but we
# only want to export the classes derived from FilterChain, so this metalcass
# grabs attributes that are missing from FilterChain from it's 'filterclass'
# attribute.
class _forward_attrs(type):
    def __getattr__(cls, name):
        attr = getattr(cls.filterclass, name)
        setattr(cls, name, attr)
        return attr


class FilterChain(metaclass=_forward_attrs):
    """One or more filters combined with AND and OR operators"""

    filterclass = NotImplemented

    def __init__(self, filters=''):
        if not isinstance(self.filterclass, type) or not issubclass(self.filterclass, Filter):
            raise RuntimeError('Attribute "filterclass" must be set to a Filter subclass')

        if isinstance(filters, str):  # Because str is also instance of abc.Sequence
            pass
        elif isinstance(filters, abc.Sequence) and all(isinstance(f, str) for f in filters):
            filters = '|'.join(filters)
        elif isinstance(filters, (type(self), self.filterclass)):
            filters = str(filters)
        elif not isinstance(filters, str):
            raise ValueError('Filters must be string or sequence of strings, not %s: %r'
                             % (type(filters).__name__, filters))

        self._filterchains = ()

        # Split `filters` at boolean operators
        parts = cliparser.tokenize(filters, delims=('&', '|'))
        if len(parts) > 0 and parts[0]:
            if parts[0] in ('&', '|'):
                raise ValueError("Filter can't start with operator: %r" % (parts[0],))
            elif parts[-1] in ('&', '|'):
                raise ValueError("Filter can't end with operator: %r" % (parts[-1],))

            # The filter chain is represented by a tuple of tuples.  Each inner
            # tuple combines filters with AND.  The outer tuple combines the
            # inner tuples with OR.
            filters = []
            ops = []
            expect = 'filter'
            for i,part in enumerate(parts):
                if expect == 'filter':
                    if part not in '&|':
                        f = self.filterclass(part)
                        if f.match_everything:
                            # One catch-all filter is the same as no filters
                            filters = [f]
                            ops.clear()
                            break
                        else:
                            filters.append(f)
                            expect = 'operator'
                            continue
                elif expect == 'operator' and part in '&|':
                    if part in '&|':
                        ops.append(part)
                        expect = 'filter'
                        continue
                raise ValueError('Consecutive operators: {!r}'.format(''.join(parts[i - 2 : i + 2])))

            fchain = [[]]
            for filter,op in itertools.zip_longest(filters, ops):
                fchain[-1].append(filter)
                if op == '|':
                    fchain.append([])
            log.debug('Chained %r and %r to %r', filters, ops, fchain)
            self._filterchains = tuple(tuple(x) for x in fchain)

    def apply(self, objects):
        """Yield matching objects from iterable `objects`"""
        chains = self._filterchains
        if chains:
            for obj in objects:
                if any(all(f.match(obj) for f in AND_chain)
                       for AND_chain in chains):
                    yield obj
        else:
            yield from objects

    def match(self, obj):
        """Whether `obj` matches this filter chain"""
        # All filters in an AND_chain must match for the AND_chain to
        # match.  At least one AND_chain must match.
        chains = self._filterchains
        if not chains:
            return True
        else:
            return any(all(f.match(obj) for f in AND_chain)
                       for AND_chain in chains)

    @property
    def needed_keys(self):
        """The object keys needed for filtering"""
        keys = set()
        for chain in self._filterchains:
            for filter in chain:
                keys.update(filter.needed_keys)
        return tuple(keys)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        else:
            # Compare sets because order doesn't matter (foo&bar|baz is the
            # same as baz|bar&foo). Use frozensets because sets are not
            # hashable.
            self_fc_sets = set(frozenset(x) for x in self._filterchains)
            other_fc_sets = set(frozenset(x) for x in other._filterchains)
            return self_fc_sets == other_fc_sets

    def __str__(self):
        if len(self._filterchains) < 1:
            return ''
        else:
            OR_chains = []
            for AND_chain in self._filterchains:
                OR_chains.append('&'.join(str(f) for f in AND_chain))
            return '|'.join(OR_chains)

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, str(self))

    def __and__(self, other):
        cls = type(self)
        if not isinstance(other, cls):
            return NotImplemented
        else:
            return cls(str(self) + '&' + str(other))

    def __or__(self, other):
        cls = type(self)
        if not isinstance(other, cls):
            return NotImplemented
        else:
            return cls(str(self) + '|' + str(other))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Base classes for documentation commands"""

from ... import __appname__, __version__, objects
from ...completion import candidates
from ...logging import make_logger
from .. import CmdError, CommandMeta

log = make_logger(__name__)



class HelpCmdbase(metaclass=CommandMeta):
    name = 'help'
    aliases = ('man',)
    category = 'miscellaneous'
    provides = set()
    description = 'List or explain commands and settings'
    usage = ('help',
             'help <TOPIC> <TOPIC> ...')
    examples = ('help',
                'help help')
    argspecs = (
        {'names': ('TOPIC',), 'nargs': '*',
         'description': 'Name of command, setting or category'},
    )

    TOPIC_DELIMITER = ('', '-  ' * 20, '')

    def run(self, TOPIC):
        topics = TOPIC
        lines = []
        success = True
        existing_topics = []

        if len(topics) < 1:
            lines = objects.helpmgr.find('overview')
        else:
            for topic in topics:
                try:
                    line = objects.helpmgr.find(topic)
                except ValueError as e:
                    self.error(e)
                    success = False
                else:
                    lines.extend(line)
                    lines.extend(self.TOPIC_DELIMITER)
                    existing_topics.append(topic)

            if lines:
                # Remove last TOPIC_DELIMITER
                for _ in range(len(self.TOPIC_DELIMITER)):
                    lines.pop(-1)

        if not existing_topics:
            existing_topics.append(__appname__)

        self.display_help(existing_topics, lines)
        if not success:
            raise CmdError()

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        return candidates.help_topics()


class VersionCmdbase(metaclass=CommandMeta):
    name = 'version'
    category = 'miscellaneous'
    provides = set()
    description = 'Show {} version'.format(__appname__)

    def run(self):
        log.info('%s version %s' % (__appname__, __version__))


class LogCmdbase(metaclass=CommandMeta):
    name = 'log'
    provides = set()
    category = 'miscellaneous'
    description = 'Clear, add or scroll through log messages'
    usage = ('log <ACTION> [<PARAMETER> <PARAMETER> ...]',)
    examples = ('log clear',
                'log scroll up',
                'log scroll page down',
                'log error Holy crap, Batman!')
    argspecs = (
        {'names': ('ACTION',), 'nargs': 'REMAINDER',
         'description': ('"clear", "scroll", "info" or "error" '
                         '(see the sections below for more information)')},
    )

    more_sections = {'clear': ('Remove all previously logged messages in the TUI.  '
                               'This action ignores all PARAMETERs.',),
                     'scroll': ('Scroll the log messages up or down in the TUI.  '
                                'Valid PARAMETERs are "up", "down", "page up", "page down", '
                                '"top" and "bottom".',),
                     'info': ('Join all PARAMETERs and display them as a normal message.',),
                     'error': ('Join all PARAMETERs and display them as an error message.',)}

    def run(self, ACTION):
        if len(ACTION) < 1:
            raise CmdError('Missing at least one argument')
        elif ACTION[0] == 'clear':
            return self._do('clear', *ACTION[1:])
        elif ACTION[0] == 'scroll':
            return self._do('scroll', *ACTION[1:])
        elif ACTION[0] == 'error':
            self.error(' '.join(ACTION[1:]))
        elif ACTION[0] == 'info':
            self.info(' '.join(ACTION[1:]))
        else:
            self.info(' '.join(ACTION))

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        if args.curarg_index == 1:
            return candidates.Candidates(
                (candidates.Candidate(topic, Description=''.join(cls.more_sections[topic]).split('.')[0])
                 for topic in cls.more_sections),
                label='Action')
        elif args.curarg_index == 2 and args[1] == 'scroll':
            return candidates.Candidates(
                (candidates.Candidate('up', Description='Scroll log messages up one line'),
                 candidates.Candidate('down', Description='Scroll log messages down one line'),
                 candidates.Candidate('page up', Description='Scroll log messages up one page'),
                 candidates.Candidate('page down', Description='Scroll log messages down one page'),
                 candidates.Candidate('top', Description='Scroll to top of log messages'),
                 candidates.Candidate('bottom', Description='Scroll to bottom of log messages')),
                label='Action')



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import os
from wcwidth import wcswidth
from unicodedata import east_asian_width as _east_asian_width
from unicodedata import normalize as _normalize_unicode


def normalize_unicode(string):
    """
    Convert combining character sequences into graphemes (e.g. "a˚" -> "å") so
    that len() reports the correct length.
    http://www.unicode.org/faq/char_combmark.html
    """
    return _normalize_unicode('NFC', string)


def striplines(lines):
    """Remove empty strings from start and end of `lines` using `pop`"""
    lines = list(lines)
    while lines and lines[0].strip() == '':
        lines.pop(0)
    while lines and lines[-1].strip() == '':
        lines.pop(-1)
    yield from lines


def wc_ljust(text, length, padding=' '):
    """Same as str.ljust() but consider zero-width and double-width characters"""
    return text + padding * max(0, (length - wcswidth(text)))


def strwidth(string):
    """Return displayed width of `string`, considering wide and zero-width characters"""
    # wcswidth() returns -1 for control characters (e.g. \n).
    return max(0, wcswidth(string))


def strcrop(string, width, tail=None):
    """Return `string` cropped to `width`, considering wide characters

    If `tail` is not None, it must be a string that is appended to the cropped
    string.
    """
    def widechar_indexes(s):
        for i,c in enumerate(s):
            if _east_asian_width(c) in 'FW':
                yield i

    if strwidth(string) <= width:
        return string  # string is already short enough

    if tail is not None:
        width -= strwidth(tail)  # Account for tail in final width

    indexes = list(widechar_indexes(string)) + [len(string)]
    if not indexes:
        return string[:width]  # No wide chars, regular cropping is ok

    parts = []
    start = 0
    end = 0
    currwidth = strwidth(''.join(parts))

    while indexes and currwidth < width and end < len(string):
        end = indexes.pop(0)
        if end > 0:
            parts.append(string[start:end])
            currwidth = strwidth(''.join(parts))
            start = end

    if currwidth > width:
        excess = currwidth - width
        parts[-1] = parts[-1][:-excess]

    if tail is not None:
        parts.append(tail)

    return ''.join(parts)


def stralign(string, width, side='left'):
    """Return `string` aligned to `side`, considering wide characters

    The returned string is filled up with spaces to take `width` single-spaced
    characters and cropped with `strcrop` if necessary.
    """
    fill = width - strwidth(string)
    if fill < 0:
        string = strcrop(string, width)
        fill = width - strwidth(string)

    if fill > 0:
        if side == 'left':
            string = string + ' ' * fill
        elif side == 'right':
            string = ' ' * fill + string
        else:
            raise TypeError("side argument must be 'left' or 'right', not {!r}".format(side))
    return string


def crop_and_align(string, width, align, has_wide_chars=True):
    """Make `string` exactly `width` spaces long and align it

    If `has_wide_chars` evaluates to True, the more expensive `stralign`
    function is used, which considers wide characters.

    `align` must be 'left' or 'right'.
    """
    if has_wide_chars:
        string = stralign(string, width, align)
    else:
        err = TypeError("'align' attribute must be 'left' or 'right', not %r}" % align)
        string_len = len(string)
        if string_len > width:
            if align == 'right':
                string = string[string_len - width:]
            elif align == 'left':
                string = string[:width]
            else:
                raise err
        else:
            if align == 'right':
                string = string.rjust(width)
            elif align == 'left':
                string = string.ljust(width)
            else:
                raise err
    return string


def common_substring(*strings):
    if not strings:
        return ''
    common = []
    i = 0
    while True:
        try:
            next_char_is_common = all(s[i] == strings[0][i] for s in strings)
        except IndexError:
            break
        else:
            if next_char_is_common:
                common.append(strings[0][i])
            else:
                break
        i += 1
    return ''.join(common) + '…'


def tildify(path):
    homedir = os.path.expanduser('~')
    if path.startswith(homedir):
        return '~' + path[len(homedir):]
    return path



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""More complex types for torrent data, e.g. lists of peers or a file tree"""

import os
import time
from collections import abc, defaultdict, deque

from . import utils
from .base import TorrentBase  # noqa: F401

from ..logging import make_logger  # isort:skip
log = make_logger(__name__)


class TorrentFilePriority(str):
    _MAP = {-2: (-2, 'off'), -1: (-1, 'low'), 0: (0, 'normal'), 1: (1, 'high'),
            'off': (-2, 'off'), 'low': (-1, 'low'), 'normal': (0, 'normal'), 'high': (1, 'high')}
    valid_values = ('off', 'low', 'normal', 'high')

    def __new__(cls, prio):
        try:
            number, string = cls._MAP[prio]
        except KeyError:
            raise ValueError('Invalid %s value: %r' % (cls.__name__, prio))
        else:
            obj = super().__new__(cls, string)
            obj.as_int = number
            return obj

    def __eq__(self, other):
        o_int, o_str = self._MAP.get(other, (None, None))
        if isinstance(other, int) and o_int == self.as_int:
            return True
        elif isinstance(other, str) and o_str == str(self):
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return self.as_int < int(other)

    def __gt__(self, other):
        return self.as_int > int(other)

    def __le__(self, other):
        return self.as_int <= int(other)

    def __ge__(self, other):
        return self.as_int >= int(other)

    def __int__(self):
        return self.as_int

    def __repr__(self):
        return '%s(%r)' % (type(self).__name__, str(self))

    def __hash__(self):
        return super().__hash__()

def _calc_percent(a, b):
    try:
        return a / b * 100
    except ZeroDivisionError:
        return 0

class TorrentFile(abc.Mapping):
    """Mapping that holds the values of a single file in a torrent"""

    # Distinguish subtrees from files without comparing classes everywhere
    nodetype = 'leaf'

    TYPES = {
        'id'              : None,
        'tid'             : None,
        'name'            : utils.SmartCmpStr,
        'path-absolute'   : utils.SmartCmpPath,
        'path-relative'   : utils.SmartCmpPath,
        'location'        : utils.SmartCmpPath,
        'size-total'      : utils.SizeInBytes,
        'size-downloaded' : utils.SizeInBytes,
        'is-wanted'       : bool,
        'priority'        : TorrentFilePriority,
        '%downloaded'     : utils.Percent,
    }

    _MODIFIERS = {
        'id'              : lambda raw: raw['id'],
        'tid'             : lambda raw: raw['tid'],
        'name'            : lambda raw: raw['name'],
        'path-absolute'   : lambda raw: os.path.join(raw['location'], raw['path'], raw['name']),
        'path-relative'   : lambda raw: os.path.join(raw['path'], raw['name']),
        'location'        : lambda raw: raw['location'],
        'size-total'      : lambda raw: raw['size-total'],
        'size-downloaded' : lambda raw: raw['size-downloaded'],
        'is-wanted'       : lambda raw: raw['is-wanted'],
        'priority'        : lambda raw: 'off' if not raw['is-wanted'] else raw['priority'],
        '%downloaded'     : lambda raw: _calc_percent(raw['size-downloaded'], raw['size-total']),
    }

    def __init__(self, tid, id, name, path, location, size_total, size_downloaded, is_wanted, priority):
        self._raw = {'tid': tid, 'id': id, 'name': name, 'path': path, 'location': location,
                     'is-wanted': is_wanted, 'priority': priority,
                     'size-total': size_total, 'size-downloaded': size_downloaded}
        self._raw_keys = list(self._raw.keys())
        self._raw_keys.sort()
        self._cache = {}

    def __getitem__(self, key):
        if key not in self._cache:
            val = self._MODIFIERS[key](self._raw)
            typ = self.TYPES.get(key)
            if typ is not None:
                self._cache[key] = typ(val)
            else:
                self._cache[key] = val
        return self._cache[key]

    def update(self, raw):
        self._raw.update(raw)
        cache = self._cache
        for key,new_value in raw.items():
            cached_value = cache.get(key)
            if cached_value is not None and cached_value != new_value:
                del cache[key]

        # %downloaded is never in raw because it is calculated from
        # size-downloaded and size-total
        if 'size-downloaded' in raw:
            try:
                del cache['%downloaded']
            except KeyError:
                pass

    def __repr__(self):
        return '<{} {!r}>'.format(type(self).__name__, self['name'])

    def __hash__(self):
        return hash(tuple([(self._raw[k]) for k in self._raw_keys]))

    def __iter__(self):
        return iter(self.TYPES)

    def __len__(self):
        return len(self.TYPES)


class TorrentPeer(abc.Mapping):
    TYPES = {
        'id'          : None,
        'tid'         : None,
        'tname'       : utils.SmartCmpStr,
        'tsize'       : utils.SizeInBytes,
        'ip'          : str,
        'port'        : int,
        'client'      : utils.SmartCmpStr,
        'downloaded'  : utils.SizeInBytes,
        '%downloaded' : utils.Percent,
        'rate-up'     : utils.BandwidthInBytes,
        'rate-down'   : utils.BandwidthInBytes,
        'eta'         : utils.Timedelta,
        'rate-est'    : utils.BandwidthInBytes,
    }

    _MODIFIERS = {
        'id'      : lambda p: (p['tid'], p['ip'], p['port']),
    }

    _MAX_PEER_PROGRESS_SAMPLE_AGE = 1800  # 30 minutes
    _PEER_PROGRESS_DATA = defaultdict(lambda: deque(maxlen=10))

    @classmethod
    def gc_peer_progress_data(cls):
        log.debug('Pruning peer progress data:')
        for peer_id,samples in tuple(cls._PEER_PROGRESS_DATA.items()):
            # Remove samples that are too old
            while samples and (samples[0][0] + cls._MAX_PEER_PROGRESS_SAMPLE_AGE) < time.monotonic():
                log.debug('Sample from %s is too old: %r', peer_id, samples[0])
                samples.popleft()

            # Remove samples where progress is 0.0.  A lot of the cache entries just have
            # a single sample from peers connecting for some reason but not downloading.
            if samples and samples[0][1] in (0.0, 1.0):
                samples.popleft()

            # Remove deque if there are no samples left
            if not samples:
                del cls._PEER_PROGRESS_DATA[peer_id]

        log.debug('Pruned peer progress data:')
        for peer_id,samples in tuple(cls._PEER_PROGRESS_DATA.items()):
            log.debug('%s: %s', peer_id, samples)

    @classmethod
    def _guess_peer_rate_and_eta(cls, peer_id, peer_progress, torrent_size):
        rate = 0
        if peer_progress >= 1:
            # Peer has already downloaded everything
            eta = utils.Timedelta.NOT_APPLICABLE
        else:
            eta = utils.Timedelta.UNKNOWN
            samples = cls._PEER_PROGRESS_DATA[peer_id]

            # Don't add the same progress twice
            if not samples or peer_progress != samples[-1][1]:
                samples.append((time.monotonic(), peer_progress))

            # We need at least 3 samples
            if len(samples) >= 3:
                # Use second and last sample to calculate rate.  The first sample is
                # ignored because its timestamp may be inaccurate: When we add the
                # first sample, the peer's progress is not current but the latest we
                # received, which happened likely tens of seconds ago.
                t_first, p_first = samples[1]
                t_last, p_last = samples[-1]
                p_diff = p_last - p_first
                t_diff = t_last - t_first

                # It's possible progress goes down, e.g. if a peer deletes a file
                if p_diff > 0:
                    torrent_size = int(torrent_size)  # Don't copy unit + unit prefix from torrent_size
                    size_diff = torrent_size * p_diff
                    rate = size_diff / t_diff
                    size_remaining = torrent_size - (torrent_size * peer_progress)
                    try:
                        eta = size_remaining / rate
                    except ZeroDivisionError:
                        eta = utils.Timedelta.UNKNOWN

        return rate, eta

    def __init__(self, tid, tname, tsize, ip, port, client, downloaded, pdownloaded, rate_up, rate_down):
        self._cache = {}
        self._dct = {'tid': tid, 'tname': tname, 'tsize': tsize,
                     'ip': ip, 'port': port, 'client': client,
                     'downloaded': downloaded, '%downloaded': pdownloaded,
                     'rate-up': rate_up, 'rate-down': rate_down}
        self._dct['rate-est'], self._dct['eta'] = \
            self._guess_peer_rate_and_eta(self['id'], pdownloaded / 100, tsize)

    def __getitem__(self, key):
        cache = self._cache
        value = cache.get(key)
        if value is None:
            modifier = self._MODIFIERS.get(key)
            if modifier is not None:
                val = modifier(self._dct)
            else:
                val = self._dct[key]
            typ = self.TYPES.get(key)
            if typ is not None:
                cache[key] = typ(val)
            else:
                cache[key] = val
        return cache[key]

    def clearcache(self):
        self._cache.clear()

    def __repr__(self):
        return '<{} #{}, {}>'.format(type(self).__name__, self['tid'], self['ip'])

    def __iter__(self):
        return iter(self.TYPES)

    def __len__(self):
        return len(self.TYPES)


class TrackerStatus(utils.SmartCmpStr):
    def __new__(cls, status):
        if status not in ('stopped', 'idle', 'queued', 'announcing', 'scraping'):
            raise ValueError('Invalid tracker status: %r' % status)
        else:
            return super().__new__(cls, status)

class TorrentTracker(abc.Mapping):
    TYPES = {
        'id'                 : None,
        'tid'                : int,
        'tname'              : utils.SmartCmpStr,
        'tier'               : int,

        'url-announce'       : utils.URL,
        'url-scrape'         : utils.URL,
        'domain'             : utils.SmartCmpStr,

        'status-announce'    : TrackerStatus,
        'status-scrape'      : TrackerStatus,
        'status'             : TrackerStatus,

        'error-announce'     : utils.SmartCmpStr,
        'error-scrape'       : utils.SmartCmpStr,
        'error'              : utils.SmartCmpStr,

        'count-downloads'    : utils.Count,
        'count-leeches'      : utils.Count,
        'count-seeds'        : utils.Count,

        'time-last-announce' : utils.Timestamp,
        'time-next-announce' : utils.Timestamp,
        'time-last-scrape'   : utils.Timestamp,
        'time-next-scrape'   : utils.Timestamp,
    }

    _MODIFIERS = {
        'id'     : lambda self: hash((self['tid'], self['url-announce'])),
        'domain' : lambda self: self['url-announce'].domain,
        'status' : lambda self: (self['status-scrape'] if self['status-announce'] == 'idle'
                                 else self['status-announce']),
        'error'  : lambda self: ('Announce error: %s' % self['error-announce']
                                 if self['error-announce'] else
                                 'Scrape error: %s' % self['error-scrape']
                                 if self['error-scrape'] else '')
    }

    def __init__(self, trkdict):
        self._dct = trkdict
        self._cache = {}

    def __getitem__(self, key):
        cache = self._cache
        value = cache.get(key)
        if value is None:
            modifier = self._MODIFIERS.get(key)
            if modifier is not None:
                val = modifier(self)
            else:
                val = self._dct[key]
            typ = self.TYPES.get(key)
            if typ is not None:
                cache[key] = typ(val)
            else:
                cache[key] = val
        return cache[key]

    def __repr__(self):
        return '<%s %s>' % (type(self).__name__, self['url-announce'])

    def __iter__(self):
        return iter(self.TYPES)

    def __len__(self):
        return len(self.TYPES)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""
Application-wide instances for the TUI
"""

import asyncio

import urwid

from .. import objects
from ..settings.defaults import DEFAULT_KEYMAP
from . import theme, urwidpatches  # noqa
from .group import Group
from .keymap import KeyMap
from .logger import LogWidget
from .miscwidgets import (AvailableDiskSpaceWidget, BandwidthStatusWidget,
                          ConnectionStatusWidget, KeyChainsWidget, MarkedItemsWidget,
                          QuickHelpWidget, TorrentCountersWidget)
from .tabs import TabBar, Tabs

from ..logging import make_logger  # isort:skip
log = make_logger(__name__)


urwidpatches.apply_patches()


keymap = KeyMap(callback=lambda cmd,widget: objects.cmdmgr.run_task(cmd, on_error=log.error))
for args in DEFAULT_KEYMAP:
    if args['action'][0] == '<' and args['action'][-1] == '>':
        args['action'] = keymap.mkkey(args['action'])
    keymap.bind(**args)


MAX_TAB_TITLE_WIDTH = 50

def _create_cli_widget():
    def reset_cli(cli):
        cli.reset()
        widgets.hide('cli')

    def run_cmd(cli):
        objects.cmdmgr.run_task(cli.edit_text, on_error=log.error)
        reset_cli(cli)

    from ..completion import candidates
    from .completer import Completer

    def get_candidates(args):
        log.debug('Getting candidates for %r', args)
        if args.curarg_index == 0:
            log.debug('Completing command: %r', args[0])
            return candidates.commands()
        else:
            cmdcls = objects.cmdmgr.get_cmdcls(args[0])
            if cmdcls is not None:
                log.debug('  Completing argument for %r', cmdcls.__name__)
                return cmdcls.completion_candidates(args)

    import os
    history_file = os.path.join(objects.localcfg['tui.cli.history-dir'].full_path, 'commands')
    from ..commands import OPS
    from .cli import CLIEditWidget
    return CLIEditWidget(prompt=':',
                         history_file=history_file,
                         on_cancel=reset_cli,
                         on_accept=run_cmd,
                         completer=Completer(get_candidates, operators=OPS))

def _greedy_spacer():
    return urwid.Padding(urwid.Text(''))


# Widget names start with "_" if they are hidden, e.g. they aren't mentioned in
# the output of "help tui".

topbar = Group(cls=urwid.Columns)
topbar.add(name='host',   widget=ConnectionStatusWidget(), options='pack')
topbar.add(name='_spacer', widget=urwid.AttrMap(_greedy_spacer(), 'topbar'))
topbar.add(name='help',   widget=QuickHelpWidget(), options='pack')

tabs = keymap.wrap(Tabs, context='tabs')(
    tabbar=urwid.AttrMap(TabBar(), 'tabs.unfocused')
)

bottombar = Group(cls=urwid.Columns)
bottombar.add(name='counters', widget=TorrentCountersWidget(), options='pack')
bottombar.add(name='_spacer1', widget=urwid.AttrMap(_greedy_spacer(), 'bottombar'))
bottombar.add(name='diskspace', widget=AvailableDiskSpaceWidget(), options='pack')
bottombar.add(name='_spacer2', widget=urwid.AttrMap(_greedy_spacer(), 'bottombar'))
bottombar.add(name='marked', widget=MarkedItemsWidget(), options='pack')
bottombar.add(name='_spacer3', widget=urwid.AttrMap(_greedy_spacer(), 'bottombar'))
bottombar.add(name='bandwidth', widget=BandwidthStatusWidget(), options='pack')

cli = urwid.AttrMap(_create_cli_widget(), 'cli')

logwidget = LogWidget(height=int(objects.localcfg['tui.log.height']),
                      autohide_delay=objects.localcfg['tui.log.autohide'])

keychains = KeyChainsWidget()
keymap.on_keychain(keychains.update)

widgets = keymap.wrap(Group, context='main')(cls=urwid.Pile)
widgets.add(name='topbar', widget=topbar, options='pack')
widgets.add(name='main', widget=tabs)
widgets.add(name='log', widget=logwidget, options='pack', visible=False)
widgets.add(name='cli', widget=cli, options='pack', visible=False)
widgets.add(name='_keychains', widget=keychains, options='pack')
widgets.add(name='bottombar', widget=bottombar, options='pack')


def unhandled_input(key):
    key = keymap.evaluate(key)
    if key is not None:
        log.debug('Unhandled key: %s', key)

urwidscreen = urwid.raw_display.Screen()
urwidloop = urwid.MainLoop(widgets,
                           screen=urwidscreen,
                           event_loop=urwid.AsyncioEventLoop(loop=asyncio.get_event_loop()),
                           unhandled_input=unhandled_input,
                           handle_mouse=False)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Column classes that display information in peer lists"""

import urwid

from ...views.peer import COLUMNS as _COLUMNS
from ..table import ColumnHeaderWidget
from .base import CellWidgetBase, Style

TUICOLUMNS = {}

class Torrent(_COLUMNS['torrent'], CellWidgetBase):
    width = ('weight', 100)
    style = Style(prefix='peerlist.torrent', focusable=False,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['torrent'].header),
                           style.attrs('header'))

TUICOLUMNS['torrent'] = Torrent


class Client(_COLUMNS['client'], CellWidgetBase):
    width = 30
    style = Style(prefix='peerlist.client', focusable=False,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['client'].header),
                           style.attrs('header'))

TUICOLUMNS['client'] = Client


class Host(_COLUMNS['host'], CellWidgetBase):
    width = ('weight', 50)
    style = Style(prefix='peerlist.host', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['host'].header),
                           style.attrs('header'))

    def update(self, data):
        if not self.text.text:
            # Set raw IP address initially
            super().update(data)

        # Set hostname via callback
        from ...objects import localcfg
        if localcfg['reverse-dns']:
            def set_hostname(hostname):
                if self.text.text != hostname:
                    self.text.set_text(hostname)
            from ...client import rdns
            rdns.query(data['ip'], callback=set_hostname)

TUICOLUMNS['host'] = Host


class Port(_COLUMNS['port'], CellWidgetBase):
    style = Style(prefix='peerlist.port', focusable=False,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['port'].header),
                           style.attrs('header'))

TUICOLUMNS['port'] = Port


class PercentDownloaded(_COLUMNS['%downloaded'], CellWidgetBase):
    style = Style(prefix='peerlist.%downloaded', focusable=False,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['%downloaded'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value < 100 else None

TUICOLUMNS['%downloaded'] = PercentDownloaded


class RateUp(_COLUMNS['rate-up'], CellWidgetBase):
    style = Style(prefix='peerlist.rate-up', focusable=False,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['rate-up'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value > 0 else None

TUICOLUMNS['rate-up'] = RateUp


class RateDown(_COLUMNS['rate-down'], CellWidgetBase):
    style = Style(prefix='peerlist.rate-down', focusable=False,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['rate-down'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value > 0 else None

TUICOLUMNS['rate-down'] = RateDown


class ETA(_COLUMNS['eta'], CellWidgetBase):
    style = Style(prefix='peerlist.eta', focusable=False,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['eta'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if bool(self.value) else None

TUICOLUMNS['eta'] = ETA


class RateEst(_COLUMNS['rate-est'], CellWidgetBase):
    style = Style(prefix='peerlist.rate-est', focusable=False,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['rate-est'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value > 0 else None

TUICOLUMNS['rate-est'] = RateEst



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import urwid
from wcwidth import wcswidth

from ... import client
from ...utils.string import wc_ljust
from ...views.torrent import COLUMNS as _COLUMNS
from ..table import ColumnHeaderWidget
from .base import CellWidgetBase, Style
from .common_columns import MarkedBase

TUICOLUMNS = {}


class Marked(MarkedBase):
    style = Style(prefix='torrentlist.marked', focusable=True, extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(), style.attrs('header'))

TUICOLUMNS['marked'] = Marked


class Id(_COLUMNS['id'], CellWidgetBase):
    style = Style(prefix='torrentlist.id', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['id'].header),
                           style.attrs('header'))

TUICOLUMNS['id'] = Id


class Infohash(_COLUMNS['infohash'], CellWidgetBase):
    style = Style(prefix='torrentlist.infohash', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['infohash'].header),
                           style.attrs('header'))

TUICOLUMNS['infohash'] = Infohash


class Name(_COLUMNS['name'], CellWidgetBase):
    width = ('weight', 100)
    style = Style(prefix='torrentlist.name', focusable=True, extras=('header',),
                  modes=('idle.progress1', 'idle.progress2', 'idle.complete',
                         'seeding.progress1', 'seeding.progress2', 'seeding.complete',
                         'stopped.progress1', 'stopped.progress2', 'stopped.complete',
                         'isolated.progress1', 'isolated.progress2', 'isolated.complete',
                         'discovering.progress1', 'discovering.progress2', 'discovering.complete',
                         'verifying.progress1', 'verifying.progress2', 'verifying.complete',
                         'downloading.progress1', 'downloading.progress2', 'downloading.complete',
                         'uploading.progress1', 'uploading.progress2', 'uploading.complete',
                         'queued.progress1', 'queued.progress2', 'queued.complete',
                         'connected.progress1', 'connected.progress2', 'connected.complete'))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['name'].header),
                           style.attrs('header'))
    needed_keys = ('name', 'status', '%downloaded', '%verified', '%metadata',
                   'rate-up', 'rate-down', 'peers-connected')

    def __init__(self, *args, **kwargs):
        self.status = ('', 'idle', 0)
        super().__init__(*args, **kwargs)

    def update(self, torrent):
        progress = torrent['%downloaded']
        Status = type(torrent['status'])
        if Status.STOPPED in torrent['status']:
            mode = 'stopped'
        elif Status.ISOLATED in torrent['status']:
            mode = 'isolated'
        elif Status.INIT in torrent['status']:
            mode = 'discovering'
            progress = torrent['%metadata']
        elif Status.VERIFY in torrent['status']:
            mode = 'verifying'
            progress = torrent['%verified']
        elif Status.DOWNLOAD in torrent['status']:
            mode = 'downloading'
        elif Status.UPLOAD in torrent['status']:
            mode = 'uploading'
        elif Status.QUEUED in torrent['status']:
            mode = 'queued'
        elif Status.CONNECTED in torrent['status']:
            mode = 'connected'
        elif Status.SEED in torrent['status']:
            mode = 'seeding'
        else:
            mode = 'idle'
        new_status = (torrent['name'], mode, progress)
        if new_status != self.status:
            self.status = new_status
            self._invalidate()

    def render(self, size, focus=False):
        (maxcol,) = size
        name, mode, progress = self.status
        # Expand name to full width
        name = wc_ljust(name, maxcol, ' ')
        if progress == 100:
            attrs = self.style.attrs(mode + '.complete', focused=focus)
            self.text.set_text((attrs, name))
        else:
            # Calculate width of left/underlined part of progress bar
            split_index = int(maxcol * progress / 100)
            # Consider characters with zero-width (e.g. U+200B ZERO WIDTH SPACE) and double-width
            # (e.g. 中)
            split_index = split_index + (len(name[:split_index]) - wcswidth(name[:split_index]))
            name1 = name[:split_index]
            name2 = name[split_index:]
            attrs1 = self.style.attrs(mode + '.progress1', focused=focus)
            attrs2 = self.style.attrs(mode + '.progress2', focused=focus)
            self.text.set_text([(attrs1, name1), (attrs2, name2)])
        return super().render(size, focus)

    def get_mode(self):
        return self.status[1]

TUICOLUMNS['name'] = Name


class Path(_COLUMNS['path'], CellWidgetBase):
    width = 20
    style = Style(prefix='torrentlist.path', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['path'].header),
                           style.attrs('header'))

TUICOLUMNS['path'] = Path


class Status(_COLUMNS['status'], CellWidgetBase):
    style = Style(prefix='torrentlist.status', focusable=True,
                  extras=('header',),
                  modes=('idle', 'downloading', 'uploading', 'connected', 'seeding',
                         'stopped', 'queued', 'isolated', 'verifying', 'discovering'))

    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['status'].header),
                           style.attrs('header'))

    MODE_MAP = {
        client.utils.Status.IDLE      : 'idle',
        client.utils.Status.DOWNLOAD  : 'downloading',
        client.utils.Status.UPLOAD    : 'uploading',
        client.utils.Status.CONNECTED : 'connected',
        client.utils.Status.SEED      : 'seeding',
        client.utils.Status.STOPPED   : 'stopped',
        client.utils.Status.QUEUED    : 'queued',
        client.utils.Status.ISOLATED  : 'isolated',
        client.utils.Status.VERIFY    : 'verifying',
        client.utils.Status.INIT      : 'discovering',
    }

    def get_mode(self):
        return self.MODE_MAP[self.value]

TUICOLUMNS['status'] = Status


class Error(_COLUMNS['error'], CellWidgetBase):
    style = Style(prefix='torrentlist.error', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['error'].header),
                           style.attrs('header'))

TUICOLUMNS['error'] = Error


class Uploaded(_COLUMNS['uploaded'], CellWidgetBase):
    style = Style(prefix='torrentlist.uploaded', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['uploaded'].header),
                           style.attrs('header'))

    def get_mode(self):
        if self.data['size-uploaded'] < self.data['size-downloaded']:
            return 'highlighted'

TUICOLUMNS['uploaded'] = Uploaded


class Downloaded(_COLUMNS['downloaded'], CellWidgetBase):
    style = Style(prefix='torrentlist.downloaded', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['downloaded'].header),
                           style.attrs('header'))

    def get_mode(self):
        t = self.data
        if t['%downloaded'] < 100 and 0 < t['size-downloaded'] < t['size-final']:
            return 'highlighted'

TUICOLUMNS['downloaded'] = Downloaded


class PercentDownloaded(_COLUMNS['%downloaded'], CellWidgetBase):
    style = Style(prefix='torrentlist.%downloaded', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['%downloaded'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value < 100 else ''

TUICOLUMNS['%downloaded'] = PercentDownloaded


class Available(_COLUMNS['available'], CellWidgetBase):
    style = Style(prefix='torrentlist.available', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['available'].header),
                           style.attrs('header'))

    def get_mode(self):
        if self.value < self.data['size-final']:
            return 'highlighted'

TUICOLUMNS['available'] = Available


class PercentAvailable(_COLUMNS['%available'], CellWidgetBase):
    style = Style(prefix='torrentlist.%available', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['%available'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value < 100 else ''

TUICOLUMNS['%available'] = PercentAvailable


class Size(_COLUMNS['size'], CellWidgetBase):
    style = Style(prefix='torrentlist.size', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['size'].header),
                           style.attrs('header'))

TUICOLUMNS['size'] = Size


class Peers(_COLUMNS['peers'], CellWidgetBase):
    style = Style(prefix='torrentlist.peers', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['peers'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value > 0 else ''

TUICOLUMNS['peers'] = Peers


class Seeds(_COLUMNS['seeds'], CellWidgetBase):
    style = Style(prefix='torrentlist.seeds', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['seeds'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value < 5 else ''

TUICOLUMNS['seeds'] = Seeds


class Ratio(_COLUMNS['ratio'], CellWidgetBase):
    style = Style(prefix='torrentlist.ratio', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['ratio'].header),
                           style.attrs('header'))

    def get_mode(self):
        if 0 <= self.value < 1:
            return 'highlighted'

TUICOLUMNS['ratio'] = Ratio


class RateUp(_COLUMNS['rate-up'], CellWidgetBase):
    style = Style(prefix='torrentlist.rate-up', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['rate-up'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value > 0 else ''

TUICOLUMNS['rate-up'] = RateUp


class RateDown(_COLUMNS['rate-down'], CellWidgetBase):
    style = Style(prefix='torrentlist.rate-down', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['rate-down'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value > 0 else ''

TUICOLUMNS['rate-down'] = RateDown


class LimitRateUp(_COLUMNS['limit-rate-up'], CellWidgetBase):
    style = Style(prefix='torrentlist.limit-rate-up', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['limit-rate-up'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value < float('inf') else ''

TUICOLUMNS['limit-rate-up'] = LimitRateUp


class LimitRateDown(_COLUMNS['limit-rate-down'], CellWidgetBase):
    style = Style(prefix='torrentlist.limit-rate-down', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['limit-rate-down'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value < float('inf') else ''

TUICOLUMNS['limit-rate-down'] = LimitRateDown


class Tracker(_COLUMNS['tracker'], CellWidgetBase):
    style = Style(prefix='torrentlist.tracker', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['tracker'].header),
                           style.attrs('header'))

TUICOLUMNS['tracker'] = Tracker


class Eta(_COLUMNS['eta'], CellWidgetBase):
    style = Style(prefix='torrentlist.eta', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['eta'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if bool(self.value) else ''

TUICOLUMNS['eta'] = Eta


class Created(_COLUMNS['created'], CellWidgetBase):
    style = Style(prefix='torrentlist.created', focusable=True, extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['created'].header),
                           style.attrs('header'))

TUICOLUMNS['created'] = Created

class Added(_COLUMNS['added'], CellWidgetBase):
    style = Style(prefix='torrentlist.added', focusable=True, extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['added'].header),
                           style.attrs('header'))

TUICOLUMNS['added'] = Added

class Started(_COLUMNS['started'], CellWidgetBase):
    style = Style(prefix='torrentlist.started', focusable=True, extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['started'].header),
                           style.attrs('header'))

TUICOLUMNS['started'] = Started

class Active(_COLUMNS['activity'], CellWidgetBase):
    style = Style(prefix='torrentlist.activity', focusable=True, extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['activity'].header),
                           style.attrs('header'))

TUICOLUMNS['activity'] = Active

class Completed(_COLUMNS['completed'], CellWidgetBase):
    style = Style(prefix='torrentlist.completed', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['completed'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value.in_future else ''

TUICOLUMNS['completed'] = Completed



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import inspect
import re
from collections import abc

from ..completion import Candidates, Categories, SingleCandidate
from ..utils import cliparser

from ..logging import make_logger  # isort:skip
log = make_logger(__name__)


class Completer():
    """Parse command line and provide completion candidates and methods"""

    def __init__(self, get_candidates, operators=()):
        assert callable(get_candidates), 'Not callable: %r' % get_candidates
        self._get_candidates = get_candidates
        self._operators = operators
        self.reset()

    async def _get_candidates_wrapper(self, args):
        async def maybe_await(x):
            while inspect.isawaitable(x):
                x = await x
            return x

        cands = await maybe_await(self._get_candidates(args))
        if cands is None:
            cats = ()
        elif isinstance(cands, Candidates):
            cats = (cands,)
        elif isinstance(cands, (abc.Iterable, abc.Iterator)):
            def flatten(iters):
                # Only yield Candidates instances
                for item in iters:
                    if item is None:
                        pass
                    elif isinstance(item, Candidates):
                        yield item
                    elif isinstance(item, abc.Iterable) and not isinstance(item, str):
                        yield from flatten(item)
                    else:
                        raise RuntimeError('Not a Candidates object or iterable of Candidates objects: %r' % (item,))
            cats = tuple(flatten(cands))
        else:
            raise RuntimeError('Invalid candidates: %r' % (cands,))

        # Include current user input (this is filled in later when the current
        # command line is parsed)
        return Categories(SingleCandidate(''), *cats)

    def reset(self):
        self._tokens = ()
        self._categories = Categories()  # Tuple of Candidates objects
        self._curpos = 0
        self._curtok_index = None
        self._curtok_curpos = None

    async def update(self, cmdline, curpos):
        # log.debug('Parsing: %r', cmdline[:curpos] + '|' + cmdline[curpos:])
        tokens = cliparser.tokenize(cmdline)
        curtok_index, curtok_curpos = cliparser.get_position(tokens, curpos)
        tokens, curtok_index, curtok_curpos = cliparser.maybe_insert_empty_token(tokens, curtok_index, curtok_curpos)
        tokens, curtok_index, curtok_curpos = cliparser.avoid_delims(tokens, curtok_index, curtok_curpos)
        # log.debug('Tokens: %r', tokens)
        curcmd_tokens, curcmd_curtok_index = cliparser.get_current_cmd(tokens, curtok_index, self._operators)
        # log.debug('Current command tokens: %r', curcmd_tokens)
        # log.debug('Focused token: %r', curcmd_curtok_index)
        if curcmd_tokens is None:
            # log.debug('No current command - no candidates')
            self.reset()
        else:
            # The candidate getter gets unescaped/unquoted tokens with delimiting
            # spaces removed (i.e. "arguments" or what would appear in sys.argv).
            # The cursor position and the index of the current argument may need to
            # be adjusted.
            args = cliparser.Args.from_tokens(curcmd_tokens, curcmd_curtok_index, curtok_curpos)

            # Get all possible candidates and find matches
            self._categories = await self._get_candidates_wrapper(args)
            self._curarg_parts = {}
            for cands in self._categories.all:
                # The candidate getter may have specified custom separators for
                # the current argument to guide us when inserting a replacement,
                # e.g. paths are separated at "/" and we don't want to complete
                # the full path, just the part between two "/".
                if cands.curarg_seps:
                    # log.debug('  Separators for current argument: %r', cands.curarg_seps)
                    curarg_parts = args.curarg.separate(cands.curarg_seps, include_seps=True)
                    # log.debug('  Current argument parts: %r', curarg_parts)
                    common_prefix = curarg_parts.curarg.before_cursor
                else:
                    common_prefix = args.curarg.before_cursor
                self._curarg_parts[cands] = common_prefix

                # log.debug('Common prefix: %r', common_prefix)
                # Filter out any candidates that don't match the current argument
                cands.reduce(r'(?i)%s' % (re.escape(common_prefix),))

            # Include current user input as the first candidate so the user can
            # select it again after selecting other candidates
            self._update_current_user_input()

            # Preserve some stuff we need for re-assembling the command line
            self._tokens = tokens
            self._curpos = curpos
            self._curtok_index = curtok_index
            self._curtok_curpos = curtok_curpos

    def complete_next(self):
        """
        Fill in next completion candidate

        Return new command line and cursor position
        """
        self.categories.next()
        self._update_current_user_input()
        return self._assemble_cmdline()

    def complete_prev(self):
        """
        Select previous completion candidate

        Return new command line and cursor position
        """
        self.categories.prev()
        self._update_current_user_input()
        return self._assemble_cmdline()

    def _assemble_cmdline(self):
        """
        Apply currently selected candidate to command line

        Return new command line string and adjusted cursor position
        """
        # log.debug('Assembling %r', self._tokens)
        if not self.categories.current:
            # Return original, unmodified command line
            return ''.join(self._tokens), self._curpos
        else:
            # Split the current token as specified by the current Candidates
            # object so we can replace part of an argument, e.g. a directory in
            # a path.
            curtok = self._tokens[self._curtok_index]
            curarg_seps = self.categories.current.curarg_seps
            curtok_parts = cliparser.tokenize(curtok, delims=curarg_seps)
            curpart_index, curpart_curpos = cliparser.get_position(curtok_parts, self._curtok_curpos)
            curtok_parts, curpart_index, curpart_curpos = \
                cliparser.maybe_insert_empty_token(curtok_parts, curpart_index, curpart_curpos, curarg_seps)
            curtok_parts, curpart_index, curpart_curpos = \
                cliparser.avoid_delims(curtok_parts, curpart_index, curpart_curpos, curarg_seps)
            # log.debug('Separated current token: %r', curtok_parts)
            self._tokens[self._curtok_index : self._curtok_index + 1] = curtok_parts
            self._curtok_index += curpart_index
            self._curtok_curpos = curpart_curpos
            # log.debug('Tokens with separated argument: %r', self._tokens)
            # log.debug('New current token: %r: %r', self._curtok_index, self._tokens[self._curtok_index])
            # log.debug('New current token cursor position: %r', self._curtok_curpos)

            # Copy user-typed tokens and insert current candidate
            curcands = self.categories.current
            curcand = curcands.current
            tokens = list(self._tokens)
            curtok = self._tokens[self._curtok_index]

            if cliparser.is_escaped(curtok):
                curcand_token = cliparser.escape(curcand)
            else:
                curcand_token = cliparser.quote(curcand)
            new_curpos = self._curpos - self._curtok_curpos + len(curcand_token)
            # log.debug('Replacing %r with %r', tokens[self._curtok_index], curcand_token)
            tokens[self._curtok_index] = curcand_token
            # log.debug('New command line: %r',
            #           ''.join(tokens[:self._curtok_index]) +
            #           tokens[self._curtok_index][new_curpos:] + '|' + tokens[self._curtok_index][new_curpos:] +
            #           ''.join(tokens[self._curtok_index:]))
            return ''.join(tokens), new_curpos

    def _update_current_user_input(self):
        """ """
        if self._categories:
            user_input_cands = self._categories.all[0]
            user_input_cands.set(self.current_user_input)
            user_input_cands.curarg_seps = self._categories.current.curarg_seps
            # log.debug('Updated current user input candidate: %r', user_input_cands)

    @property
    def current_user_input(self):
        """Portion of user input that is replaced with completion candidates"""
        # First category is always current user input
        cats = self._categories[1:]
        len_cats = len(cats)
        if len_cats == 1:
            # There is only one other category of candidates besides the current
            # user input. Use the current argument separator of that other
            # category.
            # For example, if we are completing a path, the current user input
            # is only the part between two "/", the part which we are
            # completing, not the whole path.
            curcands = cats[0]

        elif len_cats > 1 and all(cats[0].curarg_seps == cat.curarg_seps for cat in cats):
            # There are multiple categories but their curarg_seps are all equal,
            # so we can use them like described above.
            curcands = cats[0]

        else:
            # Default to whatever the special current user input instance has
            # defined for curarg_seps (most likely an empty tuple).
            curcands = self._categories[0]

        return self._curarg_parts[curcands]

    @property
    def categories(self):
        """Tuple of Candidates objects for current command line"""
        return self._categories



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ... import objects
from ...completion import candidates
from .. import CmdError
from ..base import torrent as base
from . import _mixin as mixin
from ._table import TERMSIZE, print_table

from ...logging import make_logger  # isort:skip
log = make_logger(__name__)


class AddTorrentsCmd(base.AddTorrentsCmdbase,
                     mixin.make_request):
    provides = {'cli'}

    @classmethod
    def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        # Use current working directory as base
        return candidates.fs_path(args.curarg.before_cursor,
                                  base='.',
                                  glob=r'*.torrent')


class TorrentDetailsCmd(base.TorrentDetailsCmdbase,
                        mixin.make_request, mixin.select_torrents):
    provides = {'cli'}

    async def display_details(self, torrent_id):
        from ...views.details import SECTIONS
        needed_keys = set(('name',))
        for _section in SECTIONS:
            for _item in _section['items']:
                needed_keys.update(_item.needed_keys)

        response = await self.make_request(
            objects.srvapi.torrent.torrents((torrent_id,), keys=needed_keys),
            quiet=True)
        if not response.torrents:
            raise CmdError()
        else:
            torrent = response.torrents[0]

        if TERMSIZE.columns is None:
            self._machine_readable(torrent)
        else:
            self._human_readable(torrent)

    def _human_readable(self, torrent):
        from ...views.details import SECTIONS

        label_width = max(len(item.label)
                          for section in SECTIONS
                          for item in section['items'])

        for i,section in enumerate(SECTIONS):
            if i != 0:
                print()  # Newline between sections
            print('\033[1m' + section['title'].upper() + '\033[0m')
            for item in section['items']:
                print('  %s: %s' % (item.label.rjust(label_width), item.human_readable(torrent)))

    def _machine_readable(self, torrent):
        from ...views.details import SECTIONS
        for section in SECTIONS:
            for item in section['items']:
                print('%s\t%s' % (item.label.lower(), item.machine_readable(torrent)))


class ListTorrentsCmd(base.ListTorrentsCmdbase,
                      mixin.make_request, mixin.select_torrents,
                      mixin.only_supported_columns):
    provides = {'cli'}

    async def make_torrent_list(self, tfilter, sort, columns):
        from ...views.torrent import COLUMNS as TORRENT_COLUMNS

        # Remove columns that aren't supported by CLI interface (e.g. 'marked')
        columns = self.only_supported_columns(columns, TORRENT_COLUMNS)

        # Get needed keys
        if tfilter is None:
            keys = set(sort.needed_keys)
        else:
            keys = set(sort.needed_keys + tfilter.needed_keys)

        # Get wanted torrents and sort them
        for colname in columns:
            keys.update(TORRENT_COLUMNS[colname].needed_keys)
        response = await self.make_request(
            objects.srvapi.torrent.torrents(tfilter, keys=keys),
            quiet=True)
        torrents = sort.apply(response.torrents)

        # Show table of found torrents
        if torrents:
            print_table(torrents, columns, TORRENT_COLUMNS)
        else:
            raise CmdError()


class TorrentMagnetURICmd(base.TorrentMagnetURICmdbase,
                          mixin.select_torrents):
    provides = {'cli'}

    def display_uris(self, uris):
        for uri in uris:
            print(uri)


class MoveTorrentsCmd(base.MoveTorrentsCmdbase,
                      mixin.make_request, mixin.select_torrents):
    provides = {'cli'}

    @classmethod
    async def completion_candidates_posargs(cls, args):
        """Complete positional arguments"""
        if args.curarg_index == 1:
            log.debug('Getting torrent filter candidates from %r', candidates.torrent_filter)
            return await candidates.torrent_filter(args.curarg)
        elif args.curarg_index == 2:
            return candidates.fs_path(args.curarg.before_cursor,
                                      base=objects.cfg['srv.path.complete'],
                                      directories_only=True)


class RemoveTorrentsCmd(base.RemoveTorrentsCmdbase,
                        mixin.make_request, mixin.select_torrents, mixin.ask_yes_no):
    provides = {'cli'}

    async def show_list_of_hits(self, tfilter):
        import sys
        if sys.stdout.isatty():
            cmd = 'ls --sort name %s' % tfilter
            await objects.cmdmgr.run_async(cmd)

    def remove_list_of_hits(self):
        pass


class RenameCmd(base.RenameCmdbase,
                mixin.make_request, mixin.select_torrents, mixin.select_files):
    provides = {'cli'}


class StartTorrentsCmd(base.StartTorrentsCmdbase,
                       mixin.make_request, mixin.select_torrents):
    provides = {'cli'}


class StopTorrentsCmd(base.StopTorrentsCmdbase,
                      mixin.make_request, mixin.select_torrents):
    provides = {'cli'}


class VerifyTorrentsCmd(base.VerifyTorrentsCmdbase,
                        mixin.make_request, mixin.select_torrents):
    provides = {'cli'}


class LabelCmd(base.LabelCmd, mixin.make_request, mixin.select_torrents):
    provides = {'cli'}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import os

from xdg.BaseDirectory import xdg_config_home as XDG_CONFIG_HOME
from xdg.BaseDirectory import xdg_data_home as XDG_DATA_HOME

from .. import __appname__
from ..client.sorters import PeerSorter, SettingSorter, TorrentSorter, TrackerSorter
from ..utils import convert
from ..utils.usertypes import Bool, Float, Int, Option, Path, String, Tuple
from ..views import file, peer, setting, torrent, tracker

from ..logging import make_logger  # isort:skip
log = make_logger(__name__)


DEFAULT_RCFILE      = os.path.join(XDG_CONFIG_HOME, __appname__, 'rc')
DEFAULT_HISTORY_DIR = os.path.join(XDG_DATA_HOME, __appname__, 'histories')
DEFAULT_THEME_FILE  = os.path.join(os.path.dirname(__file__), 'default.theme')

DEFAULT_TAB_COMMANDS = (
    'tab ls active|!complete',
    'tab ls -c status,seeds,ratio,size,uploaded,path,name',
    'tab ls downloading -c size,downloaded,%downloaded,%available,rate-down,completed,eta,path,name',
    'tab ls uploading -c size,ratio,uploaded,rate-up,peers,seeds,tracker,path,name',
    'tab -t peers lsp -s torrent',
    'tab ls stopped -s ratio,path -c size,%downloaded,seeds,ratio,activity,completed,path,name',
    'tab ls isolated -c error,tracker,path,name -s tracker',
    'tab --focus 1',
)


def partial_sort_order(sortercls):
    # Because sort orders have aliases and can be inverted by prepending "!", we have to
    # smarten up.

    # The SortOrder class makes sort orders compare equal to their aliases with or without
    # any of the inverter chars prepended, e.g. "name" == "!n".
    class SortOrder(str):
        _invert_chars = ''.join(TorrentSorter.INVERT_CHARS)

        def __new__(cls, name, aliases):
            self = super().__new__(cls, name)
            self._aliases = tuple(aliases)
            self._hash = hash((name,) + self._aliases)
            return self

        def __eq__(self, other):
            other_ = other.lstrip(self._invert_chars)
            return (self.lstrip(self._invert_chars) == other_ or
                    any(a.lstrip(self._invert_chars) == other_ for a in self._aliases))

        # An overloaded __eq__() obligates an overloaded __hash__(), or instances won't be
        # hashable.
        def __hash__(self):
            return self._hash

        def __repr__(self):
            return '%s(%r, aliases=%r)' % (type(self).__name__, str(self), self._aliases)

    # Wrap available sort orders in a Tuple, which also must know the aliases for each
    # sort order so it can generate the syntax string for the help output.
    sort_orders = []
    aliases = {}  # Mapping (alias -> realname)
    for name,spec in sortercls.SORTSPECS.items():
        sort_orders.append(SortOrder(name, spec.aliases))
        for alias in spec.aliases:
            aliases[alias] = name
    return Tuple.partial(options=sort_orders, aliases=aliases, dedup=True)


class Bytes(Int):
    def __new__(cls, num, unit='B', **kwargs):
        return convert.size(super().__new__(cls, num, unit=unit, **kwargs), unit='byte')


def init_defaults(localcfg):
    from .. import objects

    localcfg.add('connect.host',
                 String.partial(),
                 getter=lambda: objects.srvapi.rpc.host,
                 setter=lambda v: setattr(objects.srvapi.rpc, 'host', v),
                 default='localhost',
                 description='Hostname or IP of Transmission RPC interface')
    localcfg.add('connect.port',
                 Int.partial(min=1, max=65535, prefix='none'),
                 getter=lambda: objects.srvapi.rpc.port,
                 setter=lambda v: setattr(objects.srvapi.rpc, 'port', v),
                 default=9091,
                 description='Port of Transmission RPC interface')
    localcfg.add('connect.path',
                 String.partial(),
                 getter=lambda: objects.srvapi.rpc.path,
                 setter=lambda v: setattr(objects.srvapi.rpc, 'path', v),
                 default='/transmission/rpc',
                 description='Path of Transmission RPC interface')
    localcfg.add('connect.user',
                 String.partial(),
                 getter=lambda: objects.srvapi.rpc.user,
                 setter=lambda v: setattr(objects.srvapi.rpc, 'user', v),
                 default='',
                 description='Username to use for authentication with Transmission RPC interface')
    localcfg.add('connect.password',
                 String.partial(),
                 getter=lambda: objects.srvapi.rpc.password,
                 setter=lambda v: setattr(objects.srvapi.rpc, 'password', v),
                 default='',
                 description='Password to use for authentication with Transmission RPC interface')
    localcfg.add('connect.timeout',
                 Float.partial(min=0),
                 getter=lambda: objects.srvapi.rpc.timeout,
                 setter=lambda v: setattr(objects.srvapi.rpc, 'timeout', v),
                 default=10,
                 description='Number of seconds before connecting to Transmission RPC interface fails')
    localcfg.add('connect.tls',
                 Bool.partial(),
                 getter=lambda: objects.srvapi.rpc.tls,
                 setter=lambda v: setattr(objects.srvapi.rpc, 'tls', v),
                 default='off',
                 description='Whether to connect via HTTPS to the Transmission RPC interface')
    localcfg.add('connect.url',
                 String.partial(),
                 getter=lambda: objects.srvapi.rpc.url_unsafe,
                 setter=lambda v: setattr(objects.srvapi.rpc, 'url', v),
                 default='http://localhost:9091/transmission/rpc',
                 description='URL of the Transmission RPC interface')
    localcfg.add('connect.proxy',
                 String.partial(),
                 getter=lambda: objects.srvapi.rpc.proxy,
                 setter=lambda v: setattr(objects.srvapi.rpc, 'proxy', v),
                 default='',
                 description='SOCKS5, SOCKS4 or HTTP proxy URL to tunnel RPC communication through')

    localcfg.add('columns.torrents',
                 Tuple.partial(options=torrent.COLUMNS, aliases=torrent.ALIASES),
                 default=('marked', 'size', 'downloaded', 'uploaded', 'ratio',
                          'seeds', 'peers', 'status', 'eta', '%downloaded',
                          'rate-down', 'rate-up', 'name'),
                 description='Default columns in torrent lists')
    localcfg.add('columns.peers',
                 Tuple.partial(options=peer.COLUMNS, aliases=peer.ALIASES),
                 default=('host', 'client', '%downloaded', 'rate-down',
                          'rate-up', 'rate-est', 'eta'),
                 description='Default columns in peer lists')
    localcfg.add('columns.files',
                 Tuple.partial(options=file.COLUMNS, aliases=file.ALIASES),
                 default=('marked', 'priority', '%downloaded', 'downloaded', 'size', 'name'),
                 description='Default columns in file lists')
    localcfg.add('columns.trackers',
                 Tuple.partial(options=tracker.COLUMNS, aliases=tracker.ALIASES),
                 default=('tier', 'domain', 'error', 'last-announce', 'next-announce',
                          'leeches', 'seeds', 'downloads'),
                 description='Default columns in tracker lists')
    localcfg.add('columns.settings',
                 Tuple.partial(options=setting.COLUMNS, aliases=setting.ALIASES),
                 default=('name', 'value', 'default', 'description'),
                 description='Default columns in setting lists')

    localcfg.add('remove.max-hits',
                 Int.partial(),
                 default=10,
                 description=('Maximum number of torrents to remove without extra confirmation'
                              '; negative numbers mean "unlimited"'))

    localcfg.add('reverse-dns',
                 Bool.partial(),
                 default=True,
                 description=('Whether to lookup peers\' host names'))

    localcfg.add('sort.torrents',
                 partial_sort_order(TorrentSorter),
                 default=TorrentSorter.DEFAULT_SORT,
                 description='List of sort orders in torrent lists')
    localcfg.add('sort.peers',
                 partial_sort_order(PeerSorter),
                 default=PeerSorter.DEFAULT_SORT,
                 description='List of sort orders in peer lists')
    localcfg.add('sort.trackers',
                 partial_sort_order(TrackerSorter),
                 default=TrackerSorter.DEFAULT_SORT,
                 description='List of sort orders in tracker lists')
    localcfg.add('sort.settings',
                 partial_sort_order(SettingSorter),
                 default=SettingSorter.DEFAULT_SORT,
                 description='List of sort orders in setting lists')

    localcfg.add('tui.cli.history-dir',
                 Path.partial(base=os.path.expanduser('~')),
                 default=DEFAULT_HISTORY_DIR,
                 description='Directory where histories of user input are stored')
    localcfg.add('tui.cli.history-size',
                 Int.partial(min=0),
                 default=10000,
                 description='Maximum number of lines to keep in history files')
    localcfg.add('tui.free-space.low',
                 Bytes.partial(min=0),
                 default='10GB',
                 description='Minimum amount of free space before highlighting the display')
    localcfg.add('tui.log.height',
                 Int.partial(min=1),
                 default=10,
                 description='Maximum height of the log section')
    localcfg.add('tui.log.autohide',
                 Float.partial(min=0),
                 default=10,
                 description=('If the log is hidden, show it for this many seconds '
                              'for new log entries before hiding it again'))
    localcfg.add('tui.poll',
                 Float.partial(min=0.1),
                 default=5,
                 description='Interval in seconds between TUI updates')
    localcfg.add('tui.theme',
                 Path.partial(base=os.path.dirname(DEFAULT_RCFILE)),
                 default=DEFAULT_THEME_FILE,
                 description='Path to theme file'),

    localcfg.add('unit.bandwidth',
                 Option.partial(options=('bit', 'byte'), aliases={'b': 'bit', 'B': 'byte'}),
                 default='byte',
                 description="Unit for bandwidth rates ('bit' or 'byte')")
    localcfg.add('unitprefix.bandwidth',
                 Option.partial(options=('metric', 'binary'), aliases={'m': 'metric', 'b': 'binary'}),
                 default='metric',
                 description=("Unit prefix for bandwidth rates ('metric' or 'binary')"))

    localcfg.add('unit.size',
                 Option.partial(options=('bit', 'byte'), aliases={'b': 'bit', 'B': 'byte'}),
                 default='byte',
                 description="Unit for file sizes ('bit' or 'byte')")
    localcfg.add('unitprefix.size',
                 Option.partial(options=('metric', 'binary'), aliases={'m': 'metric', 'b': 'binary'}),
                 default='binary',
                 description=("Unit prefix for file sizes ('metric' or 'binary')"))

    localcfg.add('tui.marked.on',
                 String.partial(minlen=1, maxlen=1),
                 default='●',
                 description=("Character displayed in 'marked' column for marked "
                              "list items (see 'mark' command)"))
    localcfg.add('tui.marked.off',
                 String.partial(minlen=1, maxlen=1),
                 default=' ',
                 description=("Character displayed in 'marked' column for unmarked "
                              "list items (see 'mark' command)"))



DEFAULT_KEYMAP = (
    # Some vi and emacs key translations
    {'key': 'h',      'action': '<left>'},
    {'key': 'j',      'action': '<down>'},
    {'key': 'k',      'action': '<up>'},
    {'key': 'l',      'action': '<right>'},
    {'key': 'g',      'action': '<home>'},
    {'key': 'G',      'action': '<end>'},
    {'key': 'ctrl-n', 'action': '<down>'},
    {'key': 'ctrl-p', 'action': '<up>'},
    {'key': 'ctrl-f', 'action': '<right>'},
    {'key': 'ctrl-b', 'action': '<left>'},

    # Global TUI keys
    {'context': 'main', 'key': 'q',     'action': 'quit',
     'description': 'Exit'},
    {'context': 'main', 'key': ':',     'action': 'tui show cli',
     'description': 'Open command line interface'},
    {'context': 'main', 'key': 'alt-s', 'action': 'tab set',
     'description': 'Open settings in a new tab'},
    {'context': 'main', 'key': 'alt-shift-s', 'action': 'dump',
     'description': 'Save current settings, keybindings and tabs in the default rc file'},
    {'context': 'main', 'key': 'shift-f', 'action': 'setcommand --trailing-space tab ls',
     'description': 'Find torrents in a new tab'},

    # Help
    {'context': 'main', 'key': 'F1 c', 'action': 'tab help commands',
     'description': 'Open help for commands in a new tab'},
    {'context': 'main', 'key': 'F1 s', 'action': 'tab help settings',
     'description': 'Open help for settings in a new tab'},
    {'context': 'main', 'key': 'F1 k', 'action': 'tab help keybindings',
     'description': 'Open help for keybindings in a new tab'},
    {'context': 'main', 'key': 'F1 f', 'action': 'tab help filters',
     'description': 'Open help for filters in a new tab'},
    {'context': 'main', 'key': 'F1 r', 'action': 'tab help cfgman',
     'description': 'Open help for rc files in a new tab'},
    {'context': 'main', 'key': '?',    'action': '<F1>'},

    # Hide/Show TUI elements
    {'context': 'main', 'key': 'alt-m', 'action': 'tui toggle main',
     'description': 'Show or hide main area'},
    {'context': 'main', 'key': 'alt-t', 'action': 'tui toggle topbar',
     'description': 'Show or hide top bar'},
    {'context': 'main', 'key': 'alt-b', 'action': 'tui toggle bottombar',
     'description': 'Show or hide bottom bar'},
    {'context': 'main', 'key': 'alt-l', 'action': 'tui toggle log',
     'description': 'Show or hide log messages'},

    # Log messages
    {'context': 'main', 'key': 'ctrl-l',   'action': 'log clear',
     'description': 'Remove all log messages'},
    {'context': 'main', 'key': 'alt-pgup', 'action': 'log scroll page up',
     'description': 'Scroll log messages one page up'},
    {'context': 'main', 'key': 'alt-pgdn', 'action': 'log scroll page down',
     'description': 'Scroll log messages one page down'},
    {'context': 'main', 'key': 'alt-home', 'action': 'log scroll top',
     'description': 'Scroll to top of log messages'},
    {'context': 'main', 'key': 'alt-end',  'action': 'log scroll bottom',
     'description': 'Scroll to bottom of log messages'},

    # Global bandwidth limits
    {'context': 'main', 'key': 'shift-up',    'action': 'rate --quiet dn -- +=100kB global',
     'description': 'Increase global download rate limit by 100 kilobytes'},
    {'context': 'main', 'key': 'shift-down',  'action': 'rate --quiet dn -- -=100kB global',
     'description': 'Decrease global download rate limit by 100 kilobytes'},
    {'context': 'main', 'key': 'shift-right', 'action': 'rate --quiet up -- +=100kB global',
     'description': 'Increase global upload rate limit by 100 kilobytes'},
    {'context': 'main', 'key': 'shift-left',  'action': 'rate --quiet up -- -=100kB global',
     'description': 'Decrease global upload rate limit by 100 kilobytes'},

    # Tab management
    {'context': 'tabs', 'key': 'n',     'action': 'tab',
     'description': 'Open a new empty tab'},
    {'context': 'tabs', 'key': 'd',     'action': 'tab --close',
     'description': 'Remove current tab and focus tab on the right'},
    {'context': 'tabs', 'key': 'D',     'action': 'tab --close --focus left',
     'description': 'Remove current tab and focus tab on the left'},
    {'context': 'tabs', 'key': 'alt-1', 'action': 'tab --focus 1',
     'description': 'Focus first tab'},
    {'context': 'tabs', 'key': 'alt-2', 'action': 'tab --focus 2',
     'description': 'Focus second tab'},
    {'context': 'tabs', 'key': 'alt-3', 'action': 'tab --focus 3',
     'description': 'Focus third tab'},
    {'context': 'tabs', 'key': 'alt-4', 'action': 'tab --focus 4',
     'description': 'Focus fourth tab'},
    {'context': 'tabs', 'key': 'alt-5', 'action': 'tab --focus 5',
     'description': 'Focus fifth tab'},
    {'context': 'tabs', 'key': 'alt-6', 'action': 'tab --focus 6',
     'description': 'Focus sixth tab'},
    {'context': 'tabs', 'key': 'alt-7', 'action': 'tab --focus 7',
     'description': 'Focus seventh tab'},
    {'context': 'tabs', 'key': 'alt-8', 'action': 'tab --focus 8',
     'description': 'Focus eighth tab'},
    {'context': 'tabs', 'key': 'alt-9', 'action': 'tab --focus 9',
     'description': 'Focus ninth tab'},
    {'context': 'tabs', 'key': 'alt-0', 'action': 'tab --focus 10',
     'description': 'Focus tenth tab'},
    {'context': 'tabs', 'key': 'H',     'action': 'tab --move left',
     'description': 'Move current tab to the left'},
    {'context': 'tabs', 'key': 'L',     'action': 'tab --move right',
     'description': 'Move current tab to the right'},

    # List torrents with different filters
    {'context': 'tabs', 'key': 'f a', 'action': 'tab ls active',
     'description': 'List active torrents in a new tab'},
    {'context': 'tabs', 'key': 'f A', 'action': 'tab ls !active',
     'description': 'List inactive torrents in a new tab'},
    {'context': 'tabs', 'key': 'f i', 'action': 'tab ls isolated --columns name,tracker,error',
     'description': 'List isolated torrents in a new tab'},
    {'context': 'tabs', 'key': 'f p', 'action': 'tab ls paused',
     'description': 'List paused torrents in a new tab'},
    {'context': 'tabs', 'key': 'f P', 'action': 'tab ls !paused',
     'description': 'List unpaused torrents in a new tab'},
    {'context': 'tabs', 'key': 'f c', 'action': 'tab ls complete',
     'description': 'List complete torrents in a new tab'},
    {'context': 'tabs', 'key': 'f C', 'action': 'tab ls !complete',
     'description': 'List incomplete torrents in a new tab'},
    {'context': 'tabs', 'key': 'f u', 'action': 'tab ls uploading',
     'description': 'List uploading torrents in a new tab'},
    {'context': 'tabs', 'key': 'f d', 'action': 'tab ls downloading',
     'description': 'List downloading torrents in a new tab'},
    {'context': 'tabs', 'key': 'f s', 'action': 'tab ls seeding',
     'description': 'List seeding torrents in a new tab'},
    {'context': 'tabs', 'key': 'f l', 'action': 'tab ls leeching',
     'description': 'List leeching torrents in a new tab'},
    {'context': 'tabs', 'key': 'f .', 'action': 'tab ls',
     'description': 'List all torrents in a new tab'},

    # Torrent list actions
    {'context': 'torrentlist', 'key': 's p', 'action': 'sort --add path',
     'description': TorrentSorter.SORTSPECS['path'].description},
    {'context': 'torrentlist', 'key': 's P', 'action': 'sort --add !path',
     'description': TorrentSorter.SORTSPECS['path'].description + ' (reverse)'},
    {'context': 'torrentlist', 'key': 's n', 'action': 'sort --add name',
     'description': TorrentSorter.SORTSPECS['name'].description},
    {'context': 'torrentlist', 'key': 's N', 'action': 'sort --add !name',
     'description': TorrentSorter.SORTSPECS['name'].description + ' (reverse)'},
    {'context': 'torrentlist', 'key': 's o', 'action': 'sort --add ratio',
     'description': TorrentSorter.SORTSPECS['ratio'].description},
    {'context': 'torrentlist', 'key': 's O', 'action': 'sort --add !ratio',
     'description': TorrentSorter.SORTSPECS['ratio'].description + ' (reverse)'},
    {'context': 'torrentlist', 'key': 's d', 'action': 'sort --add %downloaded',
     'description': TorrentSorter.SORTSPECS['%downloaded'].description},
    {'context': 'torrentlist', 'key': 's D', 'action': 'sort --add !%downloaded',
     'description': TorrentSorter.SORTSPECS['%downloaded'].description + ' (reverse)'},
    {'context': 'torrentlist', 'key': 's s', 'action': 'sort --add seeds',
     'description': TorrentSorter.SORTSPECS['seeds'].description},
    {'context': 'torrentlist', 'key': 's S', 'action': 'sort --add !seeds',
     'description': TorrentSorter.SORTSPECS['seeds'].description + ' (reverse)'},
    {'context': 'torrentlist', 'key': 's t', 'action': 'sort --add tracker',
     'description': TorrentSorter.SORTSPECS['tracker'].description},
    {'context': 'torrentlist', 'key': 's T', 'action': 'sort --add !tracker',
     'description': TorrentSorter.SORTSPECS['tracker'].description + ' (reverse)'},
    {'context': 'torrentlist', 'key': 's z', 'action': 'sort --add size',
     'description': TorrentSorter.SORTSPECS['size'].description},
    {'context': 'torrentlist', 'key': 's Z', 'action': 'sort --add !size',
     'description': TorrentSorter.SORTSPECS['size'].description + ' (reverse)'},

    {'context': 'torrentlist', 'key': 's m a', 'action': 'sort --add activity',
     'description': TorrentSorter.SORTSPECS['activity'].description},
    {'context': 'torrentlist', 'key': 's m A', 'action': 'sort --add !activity',
     'description': TorrentSorter.SORTSPECS['activity'].description + ' (reverse)'},
    {'context': 'torrentlist', 'key': 's m c', 'action': 'sort --add completed',
     'description': TorrentSorter.SORTSPECS['completed'].description},
    {'context': 'torrentlist', 'key': 's m C', 'action': 'sort --add !completed',
     'description': TorrentSorter.SORTSPECS['completed'].description + ' (reverse)'},
    {'context': 'torrentlist', 'key': 's m r', 'action': 'sort --add created',
     'description': TorrentSorter.SORTSPECS['created'].description},
    {'context': 'torrentlist', 'key': 's m R', 'action': 'sort --add !created',
     'description': TorrentSorter.SORTSPECS['created'].description + ' (reverse)'},
    {'context': 'torrentlist', 'key': 's m d', 'action': 'sort --add added',
     'description': TorrentSorter.SORTSPECS['added'].description},
    {'context': 'torrentlist', 'key': 's m D', 'action': 'sort --add !added',
     'description': TorrentSorter.SORTSPECS['added'].description + ' (reverse)'},
    {'context': 'torrentlist', 'key': 's m e', 'action': 'sort --add eta',
     'description': TorrentSorter.SORTSPECS['eta'].description},
    {'context': 'torrentlist', 'key': 's m E', 'action': 'sort --add !eta',
     'description': TorrentSorter.SORTSPECS['eta'].description + ' (reverse)'},
    {'context': 'torrentlist', 'key': 's m s', 'action': 'sort --add started',
     'description': TorrentSorter.SORTSPECS['started'].description},
    {'context': 'torrentlist', 'key': 's m S', 'action': 'sort --add !started',
     'description': TorrentSorter.SORTSPECS['started'].description + ' (reverse)'},

    {'context': 'torrentlist', 'key': 's r d', 'action': 'sort --add rate-down',
     'description': TorrentSorter.SORTSPECS['rate-down'].description},
    {'context': 'torrentlist', 'key': 's r D', 'action': 'sort --add !rate-down',
     'description': TorrentSorter.SORTSPECS['rate-down'].description + ' (reverse)'},
    {'context': 'torrentlist', 'key': 's r u', 'action': 'sort --add rate-up',
     'description': TorrentSorter.SORTSPECS['rate-up'].description},
    {'context': 'torrentlist', 'key': 's r U', 'action': 'sort --add !rate-up',
     'description': TorrentSorter.SORTSPECS['rate-up'].description + ' (reverse)'},
    {'context': 'torrentlist', 'key': 's r r', 'action': 'sort --add rate',
     'description': TorrentSorter.SORTSPECS['rate'].description},
    {'context': 'torrentlist', 'key': 's r R', 'action': 'sort --add !rate',
     'description': TorrentSorter.SORTSPECS['rate'].description + ' (reverse)'},

    {'context': 'torrentlist', 'key': 's ,', 'action': 'sort --reset',
     'description': 'Reset to initial sort orders'},
    {'context': 'torrentlist', 'key': 's .', 'action': 'sort --none',
     'description': 'Remove all sort orders'},
    {'context': 'torrentlist', 'key': '/',   'action': ("interactive 'limit \"[]\"' --per-change "
                                                        "--on-cancel 'limit --clear' --ignore-errors"),
     'description': 'Reduce listed torrents by applying more filters'},

    # Torrent actions
    {'context': 'torrent', 'key': 't a',       'action': 'announce',
     'description': 'Announce torrent to its tracker(s)'},
    {'context': 'torrent', 'key': 't d',       'action': 'delete',
     'description': 'Remove torrent and keep any downloaded files'},
    {'context': 'torrent', 'key': 't D',       'action': 'delete --delete-files',
     'description': 'Remove torrent and all downloaded files'},
    {'context': 'torrent', 'key': 't m',       'action': 'setcommand move {location}/',
     'description': 'Move torrent'},
    {'context': 'torrent', 'key': 't n',       'action': 'setcommand rename id={id} {name}',
     'description': 'Rename torrent'},
    {'context': 'torrent', 'key': 't p',       'action': 'tab peerlist',
     'description': 'List selected or focused torrents\' peers in a new tab'},
    {'context': 'torrent', 'key': 't t',       'action': 'tab trackerlist',
     'description': 'List selected or focused torrents\' trackers in a new tab'},
    {'context': 'torrent', 'key': 't s',       'action': 'start --toggle',
     'description': 'Start or stop selected or focused torrents'},
    # Disabled until we have download queues implemented
    # {'context': 'torrent', 'key': 't S',       'action': 'start --toggle --force',
    #  'description': ''},
    {'context': 'torrent', 'key': 't v',       'action': 'verify',
     'description': 'Verify selected or focused torrents\' downloaded files'},
    {'context': 'torrent', 'key': 'enter',     'action': 'tab details',
     'description': 'Show details about focused torrent in a new tab'},
    {'context': 'torrent', 'key': 'alt-enter', 'action': 'tab filelist',
     'description': 'List selected or focused torrents\' files in a new tab'},
    {'context': 'torrent', 'key': 'space',     'action': 'mark --toggle --focus-next',
     'description': 'Mark or unmark focused torrent'},
    {'context': 'torrent', 'key': 'alt-space', 'action': 'mark --toggle --all',
     'description': 'Mark or unmark all torrents'},

    # Peer list actions
    {'context': 'peerlist', 'key': 's e', 'action': 'sort --add eta',
     'description': PeerSorter.SORTSPECS['eta'].description},
    {'context': 'peerlist', 'key': 's E', 'action': 'sort --add !eta',
     'description': PeerSorter.SORTSPECS['eta'].description + ' (reverse)'},
    {'context': 'peerlist', 'key': 's h', 'action': 'sort --add host',
     'description': PeerSorter.SORTSPECS['host'].description},
    {'context': 'peerlist', 'key': 's H', 'action': 'sort --add !host',
     'description': PeerSorter.SORTSPECS['host'].description + ' (reverse)'},
    {'context': 'peerlist', 'key': 's c', 'action': 'sort --add client',
     'description': PeerSorter.SORTSPECS['client'].description},
    {'context': 'peerlist', 'key': 's C', 'action': 'sort --add !client',
     'description': PeerSorter.SORTSPECS['client'].description + ' (reverse)'},
    {'context': 'peerlist', 'key': 's p', 'action': 'sort --add %downloaded',
     'description': PeerSorter.SORTSPECS['%downloaded'].description},
    {'context': 'peerlist', 'key': 's P', 'action': 'sort --add !%downloaded',
     'description': PeerSorter.SORTSPECS['%downloaded'].description + ' (reverse)'},

    {'context': 'peerlist', 'key': 's r u', 'action': 'sort --add rate-up',
     'description': PeerSorter.SORTSPECS['rate-up'].description},
    {'context': 'peerlist', 'key': 's r U', 'action': 'sort --add !rate-up',
     'description': PeerSorter.SORTSPECS['rate-up'].description + ' (reverse)'},
    {'context': 'peerlist', 'key': 's r d', 'action': 'sort --add rate-down',
     'description': PeerSorter.SORTSPECS['rate-down'].description},
    {'context': 'peerlist', 'key': 's r D', 'action': 'sort --add !rate-down',
     'description': PeerSorter.SORTSPECS['rate-down'].description + ' (reverse)'},
    {'context': 'peerlist', 'key': 's r r', 'action': 'sort --add rate',
     'description': PeerSorter.SORTSPECS['rate'].description},
    {'context': 'peerlist', 'key': 's r R', 'action': 'sort --add !rate',
     'description': PeerSorter.SORTSPECS['rate'].description + ' (reverse)'},
    {'context': 'peerlist', 'key': 's r e', 'action': 'sort --add rate-est',
     'description': PeerSorter.SORTSPECS['rate-est'].description},
    {'context': 'peerlist', 'key': 's r E', 'action': 'sort --add !rate-est',
     'description': PeerSorter.SORTSPECS['rate-est'].description + ' (reverse)'},

    {'context': 'peerlist', 'key': 's t', 'action': 'sort --add torrent',
     'description': PeerSorter.SORTSPECS['torrent'].description},
    {'context': 'peerlist', 'key': 's T', 'action': 'sort --add !torrent',
     'description': PeerSorter.SORTSPECS['torrent'].description + ' (reverse)'},
    {'context': 'peerlist', 'key': 's ,', 'action': 'sort --reset',
     'description': 'Reset to initial sort orders'},
    {'context': 'peerlist', 'key': 's .', 'action': 'sort --none',
     'description': 'Remove all sort orders'},
    {'context': 'peerlist', 'key': '/',   'action': ("interactive 'limit \"[]\"' --per-change "
                                                     "--on-cancel 'limit --clear' --ignore-errors"),
     'description': 'Reduce listed peers by applying more filters'},

    # File list actions
    {'context': 'filelist', 'key': '/', 'action': ("interactive 'limit \"[]\"' --per-change "
                                                   "--on-cancel 'limit --clear' --ignore-errors"),
     'description': 'Reduce listed files by applying more filters'},

    # File actions
    {'context': 'file', 'key': 'f n',       'action': 'setcommand rename {name}',
     'description': 'Rename file or directory'},
    {'context': 'file', 'key': 'f +',       'action': 'priority high',
     'description': 'Set selected or focused file\'s download priority to "high"'},
    {'context': 'file', 'key': 'f =',       'action': 'priority normal',
     'description': 'Set selected or focused file\'s download priority to "normal"'},
    {'context': 'file', 'key': 'f -',       'action': 'priority low',
     'description': 'Set selected or focused file\'s download priority to "low"'},
    {'context': 'file', 'key': 'f 0',       'action': 'priority off',
     'description': 'Don\'t download selected or focused file(s)'},
    {'context': 'file', 'key': 'space',     'action': 'mark --toggle --focus-next',
     'description': 'Mark or unmark focused file or directory'},
    {'context': 'file', 'key': 'alt-space', 'action': 'mark --toggle --all',
     'description': 'Mark or unmark all files'},

    # Tracker list actions
    {'context': 'trackerlist', 'key': 's e',   'action': 'sort --add error',
     'description': TrackerSorter.SORTSPECS['error'].description},
    {'context': 'trackerlist', 'key': 's E',   'action': 'sort --add !error',
     'description': TrackerSorter.SORTSPECS['error'].description + ' (reverse)'},
    {'context': 'trackerlist', 'key': 's s',   'action': 'sort --add status',
     'description': TrackerSorter.SORTSPECS['status'].description},
    {'context': 'trackerlist', 'key': 's S',   'action': 'sort --add !status',
     'description': TrackerSorter.SORTSPECS['status'].description + ' (reverse)'},
    {'context': 'trackerlist', 'key': 's d',   'action': 'sort --add domain',
     'description': TrackerSorter.SORTSPECS['domain'].description},
    {'context': 'trackerlist', 'key': 's D',   'action': 'sort --add !domain',
     'description': TrackerSorter.SORTSPECS['domain'].description + ' (reverse)'},
    {'context': 'trackerlist', 'key': 's t',   'action': 'sort --add torrent',
     'description': TrackerSorter.SORTSPECS['torrent'].description},
    {'context': 'trackerlist', 'key': 's T',   'action': 'sort --add !torrent',
     'description': TrackerSorter.SORTSPECS['torrent'].description + ' (reverse)'},

    {'context': 'trackerlist', 'key': 's n d',   'action': 'sort --add downloads',
     'description': TrackerSorter.SORTSPECS['downloads'].description},
    {'context': 'trackerlist', 'key': 's n D',   'action': 'sort --add !downloads',
     'description': TrackerSorter.SORTSPECS['downloads'].description + ' (reverse)'},
    {'context': 'trackerlist', 'key': 's n l',   'action': 'sort --add leeches',
     'description': TrackerSorter.SORTSPECS['leeches'].description},
    {'context': 'trackerlist', 'key': 's n L',   'action': 'sort --add !leeches',
     'description': TrackerSorter.SORTSPECS['leeches'].description + ' (reverse)'},
    {'context': 'trackerlist', 'key': 's n s',   'action': 'sort --add seeds',
     'description': TrackerSorter.SORTSPECS['seeds'].description},
    {'context': 'trackerlist', 'key': 's n S',   'action': 'sort --add !seeds',
     'description': TrackerSorter.SORTSPECS['seeds'].description + ' (reverse)'},

    {'context': 'trackerlist', 'key': 's a n', 'action': 'sort --add next-announce',
     'description': TrackerSorter.SORTSPECS['next-announce'].description},
    {'context': 'trackerlist', 'key': 's a N', 'action': 'sort --add !next-announce',
     'description': TrackerSorter.SORTSPECS['next-announce'].description + ' (reverse)'},
    {'context': 'trackerlist', 'key': 's a l', 'action': 'sort --add last-announce',
     'description': TrackerSorter.SORTSPECS['last-announce'].description},
    {'context': 'trackerlist', 'key': 's a L', 'action': 'sort --add !last-announce',
     'description': TrackerSorter.SORTSPECS['last-announce'].description + ' (reverse)'},

    {'context': 'trackerlist', 'key': 's ,',   'action': 'sort --reset',
     'description': 'Reset to initial sort orders'},
    {'context': 'trackerlist', 'key': 's .',   'action': 'sort --none',
     'description': 'Remove all sort orders'},
    {'context': 'trackerlist', 'key': '/',     'action': ("interactive 'limit \"[]\"' --per-change "
                                                          "--on-cancel 'limit --clear' --ignore-errors"),
     'description': 'Reduce listed trackers by applying more filters'},

    # Tracker actions
    {'context': 'torrent', 'key': 'enter', 'action': 'tab details',
     'description': 'Show details about torrent of focused tracker in new tab'},
    {'context': 'torrent', 'key': 'alt-enter', 'action': 'tab filelist',
     'description': 'List files of torrent of focused tracker in a new tab'},

    # Setting list actions
    {'context': 'settinglist', 'key': 's n', 'action': 'sort --add name',
     'description': SettingSorter.SORTSPECS['name'].description},
    {'context': 'settinglist', 'key': 's N', 'action': 'sort --add !name',
     'description': SettingSorter.SORTSPECS['name'].description + ' (reverse)'},
    {'context': 'settinglist', 'key': 's v', 'action': 'sort --add value',
     'description': SettingSorter.SORTSPECS['value'].description},
    {'context': 'settinglist', 'key': 's V', 'action': 'sort --add !value',
     'description': SettingSorter.SORTSPECS['value'].description + ' (reverse)'},
    {'context': 'settinglist', 'key': 's d', 'action': 'sort --add default',
     'description': SettingSorter.SORTSPECS['default'].description},
    {'context': 'settinglist', 'key': 's D', 'action': 'sort --add !default',
     'description': SettingSorter.SORTSPECS['default'].description + ' (reverse)'},
    {'context': 'settinglist', 'key': 's c', 'action': 'sort --add description',
     'description': SettingSorter.SORTSPECS['description'].description},
    {'context': 'settinglist', 'key': 's C', 'action': 'sort --add !description',
     'description': SettingSorter.SORTSPECS['description'].description + ' (reverse)'},
    {'context': 'settinglist', 'key': 's ,', 'action': 'sort --reset',
     'description': 'Reset to initial sort orders'},
    {'context': 'settinglist', 'key': 's .', 'action': 'sort --none',
     'description': 'Remove all sort orders'},
    {'context': 'settinglist', 'key': '/',   'action': ("interactive 'limit \"[]\"' --per-change "
                                                        "--on-cancel 'limit --clear' --ignore-errors"),
     'description': 'Reduce listed settings by applying more filters'},

    # Setting actions
    {'context': 'setting', 'key': 'ctrl-r', 'action': 'reset',
     'description': 'Reset setting to its default value'},

    # Help actions
    {'context': 'helptext', 'key': '/',      'action': ("interactive 'find \"[]\"' --per-change "
                                                        "--on-cancel 'find --clear' --ignore-errors"),
     'description': 'Search for string'},
    {'context': 'helptext', 'key': 'n',      'action': 'find --next',
     'description': 'Find next occurrence of search string'},
    {'context': 'helptext', 'key': 'ctrl-n', 'action': 'find --next',
     'description': 'Find next occurrence of search string'},
    {'context': 'helptext', 'key': 'N',      'action': 'find --previous',
     'description': 'Find previous occurrence of search string'},
    {'context': 'helptext', 'key': 'ctrl-p', 'action': 'find --previous',
     'description': 'Find previous occurrence of search string'},
)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from collections import abc

from . import OPS
from .cmdbase import _CommandBase


def is_op(string):
    return string in OPS


def is_cmdcls(obj):
    """Whether `obj` is a command class"""
    return isinstance(obj, type) and \
        obj is not _CommandBase and \
        issubclass(obj, _CommandBase)


def listify_args(args):
    """
    Make list from `args`

    Ensure `args` is a string and use split(',') to turn it into a list.  If
    `args` is any non-string iterable, recursively listify each item and append
    it to the final list.

    Despite the name, return a tuple.
    """
    if args is None:
        return []

    if isinstance(args, str):
        # args is a string of comma-separated list items
        args = (str(arg).strip() for arg in args.split(','))
    elif isinstance(args, abc.Iterable):
        # args is an iterable of list items; each item may be a string of
        # comma-separated items
        args = (a for arg in args
                for a in listify_args(arg))
    else:
        # args is something else
        args = (str(args).strip())
    return [arg for arg in args if arg]


def log_msgs(process, response, quiet=False):
    """
    Get messages for the user from `response` and report it to the user

    `process` must be an object with `info` and `error` methods (i.e. a command
    instance) that take a single string and report it to the user.

    `response` must be an object with the attributes `msgs` and `errors` (i.e. a
    Response instance).

    If `quiet` evaluates to True, only errors are reported.
    """
    if not quiet:
        for msg in response.msgs:
            process.info(msg)

    for error in response.errors:
        process.error(error)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import os

import urwid
from natsort import humansorted

from .. import objects, utils
from ..client import constants as const
from .main import redraw_screen

# Workaround for urwid bug: When a Text widget is initialized with an empty
# string, subsequent set_text() calls have no effect.
EMPTY_TEXT = ' '


class KeyChainsWidget(urwid.WidgetWrap):
    _sizing = frozenset(urwid.FLOW)
    _headers = ('Key Chain', 'Command', 'Description')

    def selectable(self):
        return False

    def __init__(self):
        self._pile = urwid.Pile([])
        super().__init__(urwid.AttrMap(self._pile, 'keychains'))

    def update(self, keymap, context, active_keychains, keys_given):
        active_keychains = list(active_keychains)
        self._pile.contents[:] = []
        if not active_keychains:
            return

        # Find number of keychain columns
        key_col_num = max(len(kc) for kc,action in active_keychains)

        # Find widest key for each column (each key is in its own cell)
        key_col_widths = [0] * key_col_num
        for kc,action in active_keychains:
            for colnum in range(key_col_num):
                try:
                    width = len(kc[colnum])
                except IndexError:
                    width = 0
                key_col_widths[colnum] = max(key_col_widths[colnum], width)
        # Total key chain column width is:
        # max(len(key) for each key) + (1 space between keys)
        key_col_width = sum(key_col_widths) + key_col_num - 1

        # Create list of rows
        keychain_col_width = max(len(self._headers[0]), key_col_width)
        spacer = ('pack' , urwid.Text('  '))
        rows = [
            # First row is the headers
            urwid.AttrMap(urwid.Columns([
                (keychain_col_width, urwid.Text(self._headers[0])),
                spacer,
                urwid.Text(self._headers[1]),
                spacer,
                urwid.Text(self._headers[2]),
            ]), 'keychains.header')
        ]
        next_key_index = len(keys_given)
        if len(active_keychains) > 10:
            active_keychains = self._compress_keychains(active_keychains, next_key_index)
        for kc,action in humansorted(active_keychains, key=lambda x: str(x[0])):
            row = []
            for colnum in range(key_col_num):
                colwidth = key_col_widths[colnum]
                try:
                    keytext = kc[colnum].ljust(colwidth)
                except IndexError:
                    # This keychain is shorter than the longest one
                    row.append(('pack', urwid.Text(('keychains.keys', ''.ljust(colwidth)))))
                else:
                    # Highlight the key the user needs to press to advance keychain
                    attrs = ('keychains.keys.next' if colnum == next_key_index else 'keychains.keys')
                    row.append(('pack', urwid.Text((attrs, keytext))))

                # Add space between this key cell and the next unless this is the last column
                if colnum < key_col_num - 1:
                    row.append(('pack', urwid.Text(('keychains.keys', ' '))))

            # Fill remaining space if 'Key Chain' header is longer than all key chains
            remaining_width = keychain_col_width - key_col_width
            row.append(('pack', urwid.Text(('keychains.keys', ''.ljust(remaining_width)))))

            row.append(spacer)
            row.append(urwid.AttrMap(urwid.Text(str(action)),
                                     'keychains.action'))
            row.append(spacer)
            row.append(urwid.AttrMap(urwid.Text(keymap.get_description(kc, context)),
                                     'keychains.description'))

            rows.append(urwid.Columns(row))

        options = self._pile.options('pack')
        for row in rows:
            self._pile.contents.append((row, options))

    @staticmethod
    def _compress_keychains(keychains, next_key_index):
        # Cut off any keys after next_key_index and deduplilcate the resulting
        # shortened keys.
        # Example: <a> <b> <1> command with argument 1
        #          <a> <b> <2> command with argument 2
        #          <a> <b> <3> command with argument 3
        # Becomes: <a> <b> …   command with argument …
        reduced = {}
        for kc,action in keychains:
            key = kc[:next_key_index + 1]
            if key != kc:
                subkeychains = (action for kc,action in keychains
                                if kc.startswith(key))
                action = utils.string.common_substring(*subkeychains)
                key += ('…',)
            if key not in reduced:
                reduced[key] = action
        return tuple(reduced.items())


class QuickHelpWidget(urwid.Text):
    def __init__(self):
        super().__init__('')

    def update(self):
        def get_key(cmd, contexts):
            """
            Return shortest key sequence that executes `cmd` in any context given in
            `contexts`
            """
            from .tuiobjects import keymap
            for context in contexts:
                keys = tuple(keymap.keys(lambda key,action: action.startswith(cmd), context))
                if keys:
                    return sorted(keys, key=lambda k: len(k))[0]

        def get_first_key(cmd, contexts):
            """Same as `get_key`, but return only first key of key sequence"""
            key = get_key(cmd, contexts)
            if isinstance(key, tuple):  # key is a key sequence
                return key[0]
            else:
                return key

        def maybe_add_entry(items, label, key):
            if key is not None:
                items.append([('topbar.help.space',  '   '),
                              ('topbar.help.key',    str(key)),
                              ('topbar.help.equals', ' '),
                              ('topbar.help.label',  label)])

        items = []
        maybe_add_entry(items, 'Settings', get_key('tab set', contexts=('main', None)))
        maybe_add_entry(items, 'Save', get_key('dump', contexts=('main', None)))
        maybe_add_entry(items, 'Prompt', get_key('tui show cli', contexts=('main', None)))
        maybe_add_entry(items, 'Quit', get_key('quit', contexts=('main', None)))
        maybe_add_entry(items, 'Help', get_first_key('tab help', contexts=('main', None)))
        self.set_text(items) if items else self.set_text('')


class ConnectionStatusWidget(urwid.WidgetWrap):
    def __init__(self):
        self._text = urwid.Text('Not connected')
        self._attrmap = urwid.AttrMap(self._text, 'topbar.host.disconnected')
        super().__init__(self._attrmap)
        objects.srvapi.rpc.on('connecting', self._handle_connecting)
        objects.srvapi.rpc.on('connected', self._handle_connected)
        objects.srvapi.rpc.on('disconnected', self._handle_disconnected)
        objects.srvapi.rpc.on('error', self._handle_error)

    @staticmethod
    def _connection_string(rpc):
        string = '%s:%s' % (rpc.host, int(rpc.port))
        if rpc.tls:
            string = 'https://' + string
        return string

    @redraw_screen
    def _handle_connecting(self, rpc):
        self._text.set_text('Connecting to %s' % self._connection_string(rpc))
        self._attrmap.set_attr_map({None: 'topbar.host.connecting'})

    @redraw_screen
    def _handle_connected(self, rpc):
        self._text.set_text('%s Transmission %s' % (self._connection_string(rpc), rpc.version))
        self._attrmap.set_attr_map({None: 'topbar.host.connected'})

    @redraw_screen
    def _handle_disconnected(self, rpc):
        self._text.set_text(self._connection_string(rpc))
        self._attrmap.set_attr_map({None: 'topbar.host.disconnected'})

    @redraw_screen
    def _handle_error(self, rpc, error):
        from ..client import RPCError
        if not isinstance(error, RPCError):
            self._text.set_text(str(error))  # error should also contain url
            self._attrmap.set_attr_map({None: 'topbar.host.disconnected'})


class BandwidthStatusWidget(urwid.Widget):
    _RATE_WIDTH = 6
    _SPACER_WIDTH = 1
    _TAIL_WIDTH = 5

    def __init__(self):
        self._text_up = urwid.Text('', align='right')
        self._text_up_limit = urwid.Text('')
        self._text_dn = urwid.Text('', align='right')
        self._text_dn_limit = urwid.Text('')

        spacer = urwid.AttrMap(urwid.SolidFill(' '), 'bottombar')
        self._spacer_canvas = spacer.render((self._SPACER_WIDTH, 1))

        def mkattr(text, attrsname):
            return urwid.AttrMap(urwid.Padding(text), attrsname)

        self._attr_up = mkattr(self._text_up, 'bottombar.bandwidth.up')
        self._attr_up_limit = mkattr(self._text_up_limit, 'bottombar.bandwidth.up')
        self._attr_dn = mkattr(self._text_dn, 'bottombar.bandwidth.down')
        self._attr_dn_limit = mkattr(self._text_dn_limit, 'bottombar.bandwidth.down')

        self._up_limit_width = 0
        self._dn_limit_width = 0
        self._connected = False

        objects.srvapi.status.on_update(self._update_current_rates)
        objects.srvapi.settings.on_update(self._update_rate_limits)

    @redraw_screen
    def _invalidate(self):
        super()._invalidate()

    def _update_current_rates(self, status):
        up = status.rate_up
        if up is not const.DISCONNECTED:
            up_attrs = 'bottombar.bandwidth.up.highlighted' if up > 0 else 'bottombar.bandwidth.up'
            self._text_up.set_text((up_attrs, up.without_unit))

        dn = status.rate_down
        if dn is not const.DISCONNECTED:
            dn_attrs = 'bottombar.bandwidth.down.highlighted' if dn > 0 else 'bottombar.bandwidth.down'
            self._text_dn.set_text((dn_attrs, dn.without_unit))
            self._connected = True
        else:
            self._connected = False
        self._invalidate()

    def _update_rate_limits(self, settings):
        up = settings['limit.rate.up']
        dn = settings['limit.rate.down']

        # Empty display if DISCONNECTED or UNLIMITED
        def as_str(rate):
            return ('' if rate in (const.DISCONNECTED, const.UNLIMITED)
                    else '/%s' % rate.without_unit)
        self._text_up_limit.set_text(as_str(up))
        self._text_dn_limit.set_text(as_str(dn))

        self._up_limit_width = len(self._text_up_limit.text)
        self._dn_limit_width = len(self._text_dn_limit.text)
        self._invalidate()

    def _mk_tail_canv(self, direction, icon):
        unit = {'bit': 'b', 'byte': 'B'}[objects.localcfg['unit.bandwidth']]
        text = urwid.Text('%s/s %s' % (unit, icon))
        attr_text = urwid.AttrMap(text, 'bottombar.bandwidth.%s' % direction)
        return attr_text.render((self._TAIL_WIDTH,))

    def render(self, size, focus=False):
        if not self._connected:
            size = self.pack(size)
            return urwid.AttrMap(urwid.SolidFill(' '), 'bottombar').render(size)

        canv_up_tail = self._mk_tail_canv('up', '↑')
        canv_dn_tail = self._mk_tail_canv('down', '↓')

        # List of (canvas, position, focus) tuples
        combinelist = []

        # Download
        canv_dn = self._attr_dn.render((self._RATE_WIDTH,))
        combinelist.append((canv_dn, None, False, self._RATE_WIDTH))
        if self._dn_limit_width > 0:
            canv_dn_limit = self._attr_dn_limit.render((self._dn_limit_width,))
            combinelist.append((canv_dn_limit, None, False, self._dn_limit_width))
        combinelist.append((canv_dn_tail, None, False, self._TAIL_WIDTH))

        combinelist.append((self._spacer_canvas, None, False, self._SPACER_WIDTH))

        # Upload
        canv_up = self._attr_up.render((self._RATE_WIDTH,))
        combinelist.append((canv_up, None, False, self._RATE_WIDTH))
        if self._up_limit_width > 0:
            canv_up_limit = self._attr_up_limit.render((self._up_limit_width,))
            combinelist.append((canv_up_limit, None, False, self._up_limit_width))
        combinelist.append((canv_up_tail, None, False, self._TAIL_WIDTH))

        return urwid.CanvasJoin(combinelist)

    _sizing = frozenset([urwid.FIXED])

    def rows(self, size, focus=False):
        return 1

    def pack(self, size, focus=False):
        cols = ((self._RATE_WIDTH * 2) + (self._TAIL_WIDTH * 2) +
                self._dn_limit_width + self._up_limit_width +
                self._SPACER_WIDTH)
        return (cols, 1)


class TorrentCountersWidget(urwid.WidgetWrap):
    def __init__(self):
        self._text = urwid.Text(EMPTY_TEXT)
        super().__init__(urwid.AttrMap(self._text, 'bottombar'))
        objects.srvapi.status.on_update(self._update_counters)

    def _update_counters(self, status):
        counters = status.count
        if const.DISCONNECTED in counters:
            self._text.set_text('')
            return

        parts = []
        for name in ('downloading', 'uploading', 'stopped'):
            count = getattr(counters, name)
            if count > 0:
                parts.append('%d %s' % (count, name))

        text = ['%s torrents' % counters.total]
        if parts:
            text[-1] += ': ' + ', '.join(parts)

        if counters.isolated > 0:
            text[-1] += ', '
            text.append(('bottombar.important', '%s isolated' % counters.isolated))

        self._text.set_text(text)


class AvailableDiskSpaceWidget(urwid.WidgetWrap):
    def __init__(self):
        self._text = urwid.Text(EMPTY_TEXT)
        self._cached_freespace = None
        super().__init__(urwid.AttrMap(self._text, 'bottombar'))
        objects.srvapi.freespace.on_update(self._update_diskspace)

    def _update_diskspace(self, freespace):
        self._cached_freespace = freespace
        parts = []
        for path,info in freespace.info.items():
            if isinstance(info.path, str):
                if info.error is not None:
                    errmsg = getattr(info.error, 'message', info.error)
                    text = '%s: %s' % (os.path.basename(info.path), errmsg)
                    attr = 'bottombar.important'
                    parts.append((attr, text))
                    parts.append(('bottombar', '  '))
                elif info.free is not None:
                    text = '%s %s' % (os.path.basename(info.path), info.free)
                    if info.free <= objects.localcfg['tui.free-space.low']:
                        attr = 'bottombar.important'
                    else:
                        attr = 'bottombar'
                    parts.append((attr, text))
                    parts.append(('bottombar', '  '))
        if parts:
            parts.insert(0, 'Free: ')
            self._text.set_text(parts[:-1])
        else:
            self._text.set_text('')

    def refresh(self):
        if self._cached_freespace is not None:
            self._update_diskspace(self._cached_freespace)


class MarkedItemsWidget(urwid.WidgetWrap):
    def __init__(self):
        self._text = urwid.Text(('bottombar', ' '))
        super().__init__(self._text)

    def update(self, count):
        if count > 0:
            new_text = ('bottombar.marked', ' %d marked ' % count)
        else:
            new_text = ('bottombar', EMPTY_TEXT)

        if new_text != self._text.text:
            self._text.set_text(new_text)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import asyncio
import shlex
from collections import abc
from contextlib import contextmanager
from importlib import import_module
from inspect import getmembers

from . import OPS_AND, OPS_OR, OPS_SEQ, _CommandBase, utils
from .cmdbase import CommandMeta
from .cmderror import CmdError, CmdNotFoundError

from ..logging import make_logger  # isort:skip
log = make_logger(__name__)


class CommandManager():
    def __init__(self, pre_run_hook=None, info_handler=None, error_handler=None):
        self._info_handler = info_handler
        self._error_handler = error_handler
        self.pre_run_hook = pre_run_hook
        self._cmds = {}
        self._active_interface = None
        self._ignored_calls = []

    @property
    def info_handler(self):
        return self._info_handler

    @info_handler.setter
    def info_handler(self, callback):
        assert callback is None or callable(callback), 'Not a callable: %r' % callback
        self._info_handler = callback

    @property
    def error_handler(self):
        return self._error_handler

    @error_handler.setter
    def error_handler(self, callback):
        assert callback is None or callable(callback), 'Not a callable: %r' % callback
        self._error_handler = callback

    def load_cmds_from_module(self, *modules):
        """
        Import modules and look for command classes

        Command classes must have set metaclass=CommandMeta.
        """
        for modname in modules:
            log.debug('Looking for commands in %s', modname)
            mod = import_module(modname)
            for member in getmembers(mod):
                cmdcls = member[1]
                if utils.is_cmdcls(cmdcls):
                    self.register(cmdcls)

    def register(self, cmdcls):
        """
        Add new command

        cmdcls: command class (must have CommandMeta as metaclass)
        """
        if not issubclass(cmdcls, _CommandBase):
            raise RuntimeError("{} must have CommandMeta as metaclass".format(cmdcls))
        elif not cmdcls.provides:
            raise RuntimeError('{} does not provide any interfaces'.format(cmdcls))

        # Store cmdcls for each of its supported interfaces
        for interface in cmdcls.provides:
            if interface not in self._cmds:
                self._cmds[interface] = {}
            log.debug('Registered %s command %s (%s)',
                      interface, cmdcls.name, type(cmdcls).__name__)
            self._cmds[interface][cmdcls.name] = cmdcls

    @property
    def active_interface(self):
        return self._active_interface

    @active_interface.setter
    def active_interface(self, interface):
        if interface in self._cmds or interface is None:
            self._active_interface = interface
        else:
            raise ValueError('No commands for interface {!r} registered'.format(interface))

    @contextmanager
    def temporary_active_interface(self, interface):
        """Context manager that temporarily changes the active interface"""
        orig_interface = self.active_interface
        self.active_interface = interface
        log.debug('Activating temporary interface %r', interface)
        yield
        log.debug('Deactivating temporary interface %r', interface)
        self.active_interface = orig_interface

    @property
    def active_commands(self):
        """
        Tuple of command classes for the active interface or all commands if no
        active interface is specified
        """
        if self._active_interface is None:
            return self.all_commands
        else:
            return tuple(self._cmds[self._active_interface].values())

    @property
    def all_commands(self):
        """Tuple of all command classes for all interfaces"""
        cmds = set()
        for interface,cmdnames in self._cmds.items():
            for cmdname in cmdnames:
                cmds.add(self._cmds[interface][cmdname])
        return tuple(cmds)

    def get_cmdcls(self, cmdname, interface='ACTIVE', exclusive=False):
        """
        Resolve command name `cmdname` to command class

        If `interface` is 'ACTIVE', return command class from
        `active_commands`.

        If `interface` is 'ANY', return command class from `all_commands`.

        If `interface` is anything else, it must be an existing interface and
        only a command that supports it is returned.

        If `exclusive` evaluates to True, the returned command class does not
        support any other interfaces.

        Returns None if no matching command class is registered.
        """
        if interface == 'ACTIVE':
            cmdpool = self.active_commands
        elif interface == 'ANY':
            cmdpool = self.all_commands
        elif isinstance(interface, abc.Hashable):
            try:
                cmdpool = tuple(self._cmds[interface].values())
            except KeyError:
                raise ValueError('Unknown interface: {!r}'.format(interface))
        else:
            raise RuntimeError('Interface type must be hashable: {!r}'.format(interface))

        for cmd in cmdpool:
            if cmdname in cmd.names:
                if not exclusive:
                    return cmd
                elif cmd.provides == (interface,):
                    return cmd

    def __getitem__(self, cmdname):
        cmd = self.get_cmdcls(cmdname, interface='ANY')
        if cmd is not None:
            return cmd
        else:
            raise KeyError(cmdname)

    def __contains__(self, cmdname):
        return self.get_cmdcls(cmdname, interface='ANY') is not None

    @property
    def categories(self):
        """Tuple of command categories built from commands' `category` attributes"""
        categories = set()
        for cmd in self.all_commands:
            categories.add(cmd.category)
        return tuple(sorted(categories))

    def _handle_final_process(self, process):
        if process.success in (True, None):
            return True
        return False

    def run_sync(self, commands, **kwargs):
        """
        Run `commands`, return boolean result

        Use this method in non-async code. The asyncio loop must be not running
        and not closed.

        command: See `split_cmdchain`

        Any other keyword arguments are forwarded to the command instance which
        makes them available as attributes so the 'run' method can access them
        via its instance (self).

        Returns True if all `commands` ran successfully, False otherwise.
        """
        log.debug('Running command chain synchronously: %r', commands)
        if not commands:
            return True  # No commands - no error
        for process in self._yield_from_cmdchain(commands, **kwargs):
            if not process.finished:
                process.wait_sync()
        return self._handle_final_process(process)

    async def run_async(self, commands, **kwargs):
        """Same as `run_sync` but in asynchronous contexts"""
        log.debug('Running command chain asynchronously: %r', commands)
        if not commands:
            return True  # No commands - no error
        for process in self._yield_from_cmdchain(commands, **kwargs):
            if not process.finished:
                await process.wait_async()
        return self._handle_final_process(process)

    def run_ignored_calls_sync(self, cmdname=None):
        """
        Execute any calls that were previously ignored because their interface was not active

        cmdname: Only run calls for this command or all calls if None.

        See `temporary_active_interface` for changing the active interface while running
        ignored calls.
        """
        calls_made = False
        success = True
        for call in self._get_ignored_calls(cmdname):
            success = success and self.run_sync(call)
            calls_made = True
        return calls_made and success

    async def run_ignored_calls_async(self, cmdname=None):
        """Asynchronous version of `run_ignored_calls_sync`"""
        calls_made = False
        success = True
        for call in self._get_ignored_calls(cmdname):
            success = success and await self.run_async(call)
            calls_made = True
        return calls_made and success

    def _get_ignored_calls(self, cmdname):
        for call in tuple(self._ignored_calls):
            if cmdname is None or cmdname == call[0][0]:
                yield call
                self._ignored_calls.remove(call)

    def run_task(self, commands, **kwargs):
        """Return Task that runs `run_async`"""
        log.debug('Creating command chain task: %r', commands)
        return asyncio.ensure_future(self.run_async(commands, **kwargs))

    def _yield_from_cmdchain(self, commands, **kwargs):
        try:
            cmdchain = self.split_cmdchain(commands)
        except ValueError as e:
            yield self._dummy_process(cmdname=None, exception=CmdError(e))
        else:
            prev_process_success = True
            for item in cmdchain:
                if item in OPS_AND and not prev_process_success:
                    log.debug('Found operator %s and previous command failed (%r) - aborting', item, prev_process_success)
                    break
                elif item in OPS_OR and prev_process_success:
                    log.debug('Found operator %s and previous command succeeded (%r) - aborting', item, prev_process_success)
                    break
                elif utils.is_op(item):
                    continue
                else:
                    process = self._create_process(item, **kwargs)
                    yield process
                    assert process.finished, 'Not finished: %r' % process
                    prev_process_success = process.success

    def split_cmdchain(self, commands):
        """
        Parse and validate chained commands

        commands: Command line as string or a valid command chain

        A command chain is a sequence of sequences of commands with their
        arguments.  Sub-sequences must be separated by single operators.

        Command operators are characters specified by the variables OPS_AND,
        OPS_OR and OPS_SEQ.  In a string, they must be enclosed by spaces
        (e.g. " & ").

        Example:

        >>> list(cmdmgr.split_cmdchain('do that & act "like this"'))
        [ ['do', 'that'], '&', ['act', 'like this'] ]

        Raise ValueError if `commands` is not valid.
        """
        if not isinstance(commands, abc.Sequence):
            raise RuntimeError('Must be string or sequence, not %r: %r' %
                               (type(commands).__name__, commands))
        elif isinstance(commands, str):
            args = shlex.split(commands)
            cmdchain = []
            cmd = []
            while args:
                arg = args.pop(0)
                if arg in OPS_SEQ:
                    if cmd:
                        cmdchain.append(cmd)
                    cmdchain.append(OPS_SEQ[0])
                    cmd = []
                elif arg in OPS_OR:
                    if cmd:
                        cmdchain.append(cmd)
                    cmdchain.append(OPS_OR[0])
                    cmd = []
                elif arg in OPS_AND:
                    if cmd:
                        cmdchain.append(cmd)
                    cmdchain.append(OPS_AND[0])
                    cmd = []
                else:
                    cmd.append(arg)
            if cmd:
                cmdchain.append(cmd)
        else:
            cmdchain = [str(item) if utils.is_op(item) else list(item)
                        for item in commands]

        for item in cmdchain:
            self._validate_cmdchain_item(item)

        if cmdchain and utils.is_op(cmdchain[-1]):
            cmdchain.pop(-1)

        log.debug('Parsed command chain: %r', cmdchain)
        return cmdchain

    def _create_process(self, cmdline, **kwargs):
        """Call one command and return its instance or None on error"""
        # Make a copy so we can pop from it
        cmdline = list(cmdline)

        if self.pre_run_hook is not None:
            old_cmdline = cmdline.copy()
            cmdline = self.pre_run_hook(cmdline)
            if cmdline != old_cmdline:
                log.debug('Pre-run-hook %r converted %r to %r',
                          self.pre_run_hook.__name__, old_cmdline, cmdline)

        try:
            cmdname = cmdline[0]
        except IndexError:
            # cmdline is empty
            return self._dummy_process(cmdname=None)
        cmdargs = cmdline[1:]
        cmdcls = self.get_cmdcls(cmdname, interface='ACTIVE')
        if cmdcls is None:
            if self.get_cmdcls(cmdname, interface='ANY') is not None:
                # Command exists but not in active interface.  Store it in case we want to
                # run it later and run a fake command so command chains are still working.
                log.debug('Ignoring inactive command: %s', cmdname)
                self._ignored_calls.append(((cmdname,) + tuple(cmdargs),))
                process = self._dummy_process(cmdname)
            else:
                exc = CmdNotFoundError('Unknown command')
                process = self._dummy_process(cmdname, exception=exc)
        elif cmdcls not in self.active_commands:
            exc = CmdError('%s: No support for %s interface' % (cmdname, self._active_interface))
            process = self._dummy_process(cmdname, exception=exc)
        else:
            process = cmdcls(cmdargs,
                             argv=tuple(cmdline),
                             info_handler=self._info_handler,
                             error_handler=self._error_handler,
                             **kwargs)
        return process

    def _validate_cmdchain_item(self, item):
        # item must be operator or a command line in list form
        if not (utils.is_op(item) or (isinstance(item, abc.Sequence) and not isinstance(item, str) and
                                      all(isinstance(arg, str) for arg in item))):
            raise RuntimeError('Invalid type for command chain item: %r' % (item,))

        # Test if item is an operator after another operator
        try:
            prev_item = self._prev_validation_item
        except AttributeError:
            prev_item = None
        self._prev_validation_item = item
        if utils.is_op(prev_item) and utils.is_op(item):
            raise ValueError('Consecutive operators: %s %s' % (prev_item, item))

    def _dummy_process(self, cmdname, exception=None, info=None):
        """
        Create new _CommandBase-derived class and return an instance of it

        This is useful, for example, for unknown commands, which can still be
        used in chains as a dummy process, e.g. "invalidcommand | validcommand"
        runs "validcommand" because "invalidcommand" fails.
        """
        class DummyCommand(metaclass=CommandMeta):
            name = cmdname
            category = 'dummies'
            provides = set()
            description = 'Dummy command'

            def run(self):
                if info:
                    self.info(info)
                if exception:
                    raise exception

            def info(self, msg):
                if self.name is None:
                    self._info_handler(str(msg))
                else:
                    return super().info(msg)

            def error(self, msg):
                if self.name is None:
                    self._error_handler(str(msg))
                else:
                    return super().error(msg)

        return DummyCommand((),
                            info_handler=self._info_handler,
                            error_handler=self._error_handler)
